# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from csv import DictWriter
from dataclasses import dataclass
from pathlib import Path
from time import perf_counter_ns
from typing import TYPE_CHECKING, Protocol, TextIO
from uuid import UUID, uuid4

from xopen import xopen

from lenskit.logging import get_logger
from lenskit.logging.multiprocess import Monitor, RecordSink, WorkerContext, get_monitor

if TYPE_CHECKING:
    from ._impl import Pipeline

_log = get_logger(__name__)


class ProfileSink(RecordSink[dict[str, float]], Protocol):
    def run_profiler(self) -> RunProfiler:
        return RunProfiler(self)


class PipelineProfiler:
    """
    Collect pipeline run statistics for profiling pipeline executions.
    """

    writer: DictWriter[str]
    output: TextIO
    sink_id: UUID
    _monitor: Monitor | None = None

    def __init__(self, pipeline: Pipeline, file: Path):
        self.sink_id = uuid4()
        self.output = xopen(file, "wt")
        stages = pipeline.component_names()
        _log.debug("starting pipeline profiler %s", self.sink_id)
        self.writer = DictWriter(self.output, stages)
        self.writer.writeheader()

    def set_stages(self, stages: list[str]):
        self.stages = stages
        if self.output:
            self.writer = DictWriter(self.output, stages)
            self.writer.writeheader()

    def multiprocess(self) -> ProfileSink:
        _log.debug("registering profiler %s with monitor", self.sink_id)
        self._monitor = get_monitor()
        self._monitor.add_record_sink(self)
        return WorkerProfileSink(self.sink_id)

    def close(self):
        if self._monitor is not None:
            _log.debug("waiting for monitor to quiesce")
            self._monitor.await_quiesce()

        self.output.close()
        if self._monitor is not None:
            self._monitor.remove_record_sink(self.sink_id)
            self._monitor = None

    def record(self, record: dict[str, float]):
        self.writer.writerow(record)

    def run_profiler(self):
        return RunProfiler(self)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


@dataclass
class WorkerProfileSink:
    sink_id: UUID

    def record(self, record: dict[str, float]):
        ctx = WorkerContext.active()
        if ctx is not None:
            ctx.send_record(self.sink_id, record)

    def run_profiler(self):
        return RunProfiler(self)


class RunProfiler:
    """
    Internal utility class for profiling a single pipeline run.
    """

    profiler: ProfileSink
    _record: dict[str, float]
    _starts: dict[str, float]

    def __init__(self, profiler: ProfileSink):
        self.profiler = profiler
        self._record = {}
        self._starts = {}

    def start_stage(self, stage: str):
        self._starts[stage] = perf_counter_ns()

    def finish_stage(self, stage: str):
        now = perf_counter_ns()
        start = self._starts[stage]

        self._record[stage] = now - start

    def finish_run(self):
        self.profiler.record(self._record)



#!/usr/bin/env python
# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Manipulate LensKit versions.

Usage:
    version-tool.py [-v | -vv] [-q]
    version-tool.py [-v | -vv] --github
    version-tool.py [-v | -vv] --update [DIR]

Options:
    -v, --verbose
        Ouput verbose logging.
    -q, --quiet
        Print version only, without other text.
    --github
        Write output to GitHub Actions output.
    --update
        Update version files.
"""

import logging
import os
import sys
from pathlib import Path

import tomlkit
from docopt import docopt

if prj_root := os.environ.get("MISE_PROJECT_ROOT", None):
    sys.path.insert(0, f"{prj_root}/src")

from lenskit._version import lenskit_version

_log = logging.getLogger("lenskit.version-tool")

options = docopt(__doc__ or "", options_first=True)

# set up logging
verbosity = options["--verbose"]
if verbosity > 1:
    level = logging.DEBUG
elif verbosity:
    level = logging.INFO
else:
    level = logging.WARNING
logging.basicConfig(level=level, stream=sys.stderr)

# get the version
version = lenskit_version()

if options["--github"]:
    gh_file = os.environ["GITHUB_OUTPUT"]
    with open(gh_file, "at") as ghf:
        print(f"version={version}", file=ghf)

elif options["--update"]:
    stage_dir = Path(options["DIR"] or ".")

    ppt_file = stage_dir / "pyproject.toml"
    with ppt_file.open("rt") as tf:
        meta = tomlkit.load(tf)

    proj = meta["project"]
    proj["dynamic"].remove("version")  # type: ignore
    proj["version"] = str(version)  # type: ignore

    _log.info("updating %s", str(ppt_file))
    with ppt_file.open("wt") as tf:
        _log.debug("writing updated pyproject.toml")
        tomlkit.dump(meta, tf)

else:
    # we only want to print the version
    if options["--quiet"]:
        print(version)
    else:
        print("LensKit version", version)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Relationship accessors for Dataset.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import scipy.sparse as sps
import torch
from numpy.typing import NDArray
from typing_extensions import Literal, overload, override

from lenskit._accel import data as _accel_data
from lenskit.diagnostics import FieldError
from lenskit.logging import get_logger
from lenskit.random import random_generator

from .items import ItemList
from .matrix import COOStructure, CSRStructure, SparseRowArray
from .schema import RelationshipSchema, id_col_name, num_col_name
from .types import ID, LAYOUT, MAT_AGG
from .vocab import Vocabulary

if TYPE_CHECKING:
    from .collection import ItemListCollection

_log = get_logger(__name__)


class RelationshipSet:
    """
    Representation for a set of relationship records.  This is the class for
    accessing general relationships, with arbitrarily many entity classes
    involved and repeated relationships allowed.

    For two-entity relationships without duplicates (including relationships
    formed by coalescing repeated relationships or interactions),
    :class:`MatrixRelationshipSet` extends this with additional capabilities.

    Relationship sets can be pickled or serialized, and will not save the entire
    dataset with them.  They are therefore safe to save as component elements
    during training processes.

    .. note::

        Client code does not need to construct this class; obtain instances from
        a dataset's :meth:`~lenskit.data.Dataset.relationships` or
        :meth:`~lenskit.data.Dataset.interactions` method.

    Stability:
        Caller
    """

    name: str
    """
    The name of the relationship class for these relationships.
    """
    schema: RelationshipSchema
    _matrix_set: dict[tuple[str, str], MatrixRelationshipSet]

    _table: pa.Table
    """
    The Arrow table of relationship information.
    """

    _vocabularies: dict[str, Vocabulary]
    _link_cols: list[str]

    def __init__(
        self,
        name: str,
        vocabularies: dict[str, Vocabulary],
        schema: RelationshipSchema,
        table: pa.Table,
    ):
        self.name = name
        self.schema = schema
        self._matrix_set = {}
        self._table = table

        self._vocabularies = vocabularies
        self._link_cols = [num_col_name(e) for e in schema.entities]

    def __getstate__(self):
        return {
            "name": self.name,
            "schema": self.schema,
            "columns": self._link_cols,
            "table": self._table,
            "vocabularies": self._vocabularies,
        }

    def __setstate__(self, state):
        self.name = state["name"]
        self.schema = state["schema"]
        self._link_cols = state["columns"]
        self._table = state["table"]
        self._vocabularies = state["vocabularies"]

    @property
    def is_interaction(self) -> bool:
        """
        Query whether these relationships represent interactions.
        """
        return self.schema.interaction

    @property
    def entities(self) -> list[str]:
        return [(c if e is None else e) for c, e in self.schema.entities.items()]

    @property
    def attribute_names(self) -> list[str]:
        return [c for c in self._table.column_names if c not in self._link_cols]

    def count(self):
        if "count" in self._table.column_names:
            count_column = self._table.column("count")
            count_column = count_column.fill_null(pa.scalar(1, type=pa.int8()))
            count = pc.sum(count_column)
            return count.as_py()

        return self._table.num_rows

    def arrow(self, *, attributes: str | list[str] | None = None, ids=False) -> pa.Table:
        """
        Get these relationships and their attributes as a PyArrow table.

        Args:
            attributes:
                The attributes to select.
            ids:
                If ``True``, include ID columns for the entities, instead of
                just the number columns.
        """
        table = self._table
        cols = self._link_cols

        if ids:
            id_cols = {}
            for e in self.schema.entity_class_names:
                id_cols[id_col_name(e)] = pa.array(
                    self._vocabularies[e].ids(table.column(num_col_name(e)).to_numpy())
                )
            id_tbl = pa.table(id_cols)
            cols = id_tbl.column_names
            for col in table.column_names:
                if col in self._link_cols:
                    continue

                id_tbl = id_tbl.append_column(col, table.column(col))
            table = id_tbl

        if attributes is not None:
            if isinstance(attributes, str):
                attr_cols = [attributes]
            else:
                attr_cols = attributes
            for ac in attr_cols:
                if ac not in table.column_names:
                    raise FieldError(self.name, ac)
            table = table.select(cols + attr_cols)

        return table

    def pandas(self, *, attributes: str | list[str] | None = None, ids=False) -> pd.DataFrame:
        """
        Get these relationship and their attributes as a PyArrow table.

        Args:
            attributes:
                The attributes to include in the resulting table.
            ids:
                If ``True``, include ID columns for the entities, instead of
                just the number columns.
        """
        tbl = self.arrow(attributes=attributes, ids=ids)
        return tbl.to_pandas()

    def matrix(
        self, *, row_entity: str = "user", col_entity: str = "item"
    ) -> MatrixRelationshipSet:
        """
        Convert this relationship set into a matrix, coalescing duplicate
        observations.

        Args:
            row_entity:
                The specified row entity of the matrix
            col_entity:
                The specified column entity of the matrix
        """
        mat = self._matrix_set.get((row_entity, col_entity), None)
        if mat is None:
            mat = self._make_matrix(row_entity=row_entity, col_entity=col_entity)
            self._matrix_set[(row_entity, col_entity)] = mat
        return mat

    def _make_matrix(
        self, *, row_entity: str = "user", col_entity: str = "item"
    ) -> MatrixRelationshipSet:
        if row_entity not in self.schema.entities.keys():
            raise FieldError(self.name, row_entity)

        if col_entity not in self.schema.entities.keys():
            raise FieldError(self.name, col_entity)

        if col_entity == row_entity:
            raise ValueError("row and column entity should not be the same")

        e_dict: dict[str, str | None]
        e_dict = {
            row_entity: self.schema.entities[row_entity],
            col_entity: self.schema.entities[col_entity],
        }
        matrix_schema = RelationshipSchema(
            entities=e_dict,
            interaction=self.schema.interaction,
            repeats=self.schema.repeats,
            attributes=self.schema.attributes,
        )
        new_table = self._table

        if "count" in new_table.column_names:
            null_filled = new_table["count"].fill_null(1)
            new_table = new_table.set_column(
                new_table.schema.get_field_index("count"),
                "count",
                null_filled,  # type: ignore
            )

        aggregates: list[tuple[str, str]]
        column_renames: dict[str, str]
        use_threads = True

        group_keys = [entity + "_num" for entity in e_dict]
        exclude_last = group_keys + ["timestamp", "count"]

        aggregates = [
            (col_name, "last")
            for col_name in new_table.column_names
            if col_name not in exclude_last
        ]
        column_renames = {
            f"{col_name}_last": col_name
            for col_name in new_table.column_names
            if col_name not in exclude_last
        }
        if aggregates:
            use_threads = False

        table_group = new_table.group_by(group_keys, use_threads=use_threads)

        if "timestamp" in new_table.column_names:
            aggregates.extend([("timestamp", "max"), ("timestamp", "min")])
            column_renames.update(
                {"timestamp_max": "timestamp", "timestamp_min": "first_timestamp"}
            )

        if "count" in new_table.column_names:
            aggregates.append(("count", "sum"))
            column_renames.update({"count_sum": "count"})
        else:
            aggregates.append((row_entity + "_num", "count"))
            column_renames.update({row_entity + "_num_count": "count"})

        aggregated_table = table_group.aggregate(aggregates)
        aggregated_table = aggregated_table.rename_columns(column_renames)

        return MatrixRelationshipSet(self.name, self._vocabularies, matrix_schema, aggregated_table)


class MatrixRelationshipSet(RelationshipSet):
    """
    Two-entity relationships without duplicates, accessible in matrix form.

    .. note::

        Client code does not need to construct this class; obtain instances from
        a relationship set's :meth:`~RelationshipSet.matrix` method.
    """

    _row_ptrs: np.ndarray[tuple[int], np.dtype[np.int64]]
    _structure: SparseRowArray
    row_type: str
    _row_nums: pa.Int32Array
    _row_stats: pd.DataFrame | None = None

    col_type: str
    _col_nums: pa.Int32Array
    _col_stats: pd.DataFrame | None = None

    _coords: _accel_data.CoordinateTable

    def __init__(
        self,
        name: str,
        vocabularies: dict[str, Vocabulary],
        schema: RelationshipSchema,
        table: pa.Table,
    ):
        super().__init__(name, vocabularies, schema, table)
        self._init_structures()

    def _init_structures(self, *, ds_name: str | None = None):
        log = _log.bind(dataset=ds_name, relationship=self.name)

        # order the table to compute the sparse matrix
        log.debug("setting up entity information")
        entities = list(self.schema.entities.keys())
        row, col = entities

        self.row_type = row
        self.col_type = col

        e_cols = [num_col_name(e) for e in entities]

        self._table = self._table.combine_chunks()

        # set up the coordinate table
        # TODO: make this use the cached coordinate table if possible
        self._coords = _accel_data.CoordinateTable(2)
        self._coords.extend(self._table.select(e_cols).to_batches())

        # compute the row pointers
        log.debug("computing CSR data")
        n_rows = len(self.row_vocabulary)
        row_sizes = np.zeros(n_rows + 1, dtype=np.int64)
        self._row_nums = self._table.column(e_cols[0]).combine_chunks()  # type: ignore
        rsz_struct = pc.value_counts(self._row_nums)
        rsz_nums = rsz_struct.field("values")
        rsz_counts = rsz_struct.field("counts").cast(pa.int32())
        row_sizes[np.asarray(rsz_nums) + 1] = rsz_counts
        self._row_ptrs = np.cumsum(row_sizes, dtype=np.int64)  # type: ignore

        self._col_nums = self._table.column(e_cols[1]).combine_chunks()  # type: ignore
        self._structure = SparseRowArray.from_arrays(
            self._row_ptrs,
            self._col_nums,
            shape=(len(self.row_vocabulary), len(self.col_vocabulary)),
        )

        log.debug("relationship set ready to use")

    def __getstate__(self):
        return {
            "name": self.name,
            "schema": self.schema,
            "columns": self._link_cols,
            "table": self._table,
            "vocabularies": self._vocabularies,
        }

    def __setstate__(self, state):
        self.name = state["name"]
        self.schema = state["schema"]
        self._link_cols = state["columns"]
        self._table = state["table"]
        self._vocabularies = state["vocabularies"]
        self._init_structures()

    @property
    def row_vocabulary(self):
        "The vocabulary for row entities."
        return self._vocabularies[self.row_type]

    @property
    def col_vocabulary(self):
        "The vocabulary for column entities."
        return self._vocabularies[self.col_type]

    @property
    def n_rows(self):
        return len(self.row_vocabulary)

    @property
    def n_cols(self) -> int:
        return len(self.col_vocabulary)

    @override
    def matrix(
        self, *, combine: MAT_AGG | dict[str, MAT_AGG] | None = None
    ) -> MatrixRelationshipSet:
        # already a matrix relationship set
        return self

    @overload
    def csr_structure(self, *, format: Literal["numpy"] = "numpy") -> CSRStructure: ...
    @overload
    def csr_structure(self, *, format: Literal["arrow"]) -> SparseRowArray: ...
    def csr_structure(
        self, *, format: Literal["numpy", "arrow"] = "numpy"
    ) -> CSRStructure | SparseRowArray:
        """
        Get the compressed sparse row structure of this relationship matrix.
        """
        if format == "arrow":
            return self._structure
        else:
            n_rows = len(self.row_vocabulary)
            n_cols = len(self.col_vocabulary)

            colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
            return CSRStructure(self._row_ptrs, colinds, (n_rows, n_cols))

    def coo_structure(self) -> COOStructure:
        """
        Get the compressed sparse row structure of this relationship matrix.
        """
        n_rows = len(self.row_vocabulary)
        n_cols = len(self.col_vocabulary)

        rowinds = self._table.column(num_col_name(self.row_type)).to_numpy()
        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        return COOStructure(rowinds, colinds, (n_rows, n_cols))

    @overload
    def scipy(
        self, attribute: str | None = None, *, layout: Literal["coo"], legacy: Literal[True]
    ) -> sps.coo_matrix: ...
    @overload
    def scipy(
        self,
        attribute: str | None = None,
        *,
        layout: Literal["coo"],
        legacy: Literal[False] = False,
    ) -> sps.coo_array[Any, tuple[int, int]]: ...
    @overload
    def scipy(
        self, attribute: str | None = None, *, layout: Literal["csr"] = "csr", legacy: Literal[True]
    ) -> sps.csr_matrix: ...
    @overload
    def scipy(
        self,
        attribute: str | None = None,
        *,
        layout: Literal["csr"] = "csr",
        legacy: Literal[False] = False,
    ) -> sps.csr_array[Any, tuple[int, int]]: ...
    @overload
    def scipy(
        self, attribute: str | None = None, *, layout: LAYOUT = "csr", legacy: bool = False
    ) -> sps.sparray | sps.spmatrix: ...
    def scipy(
        self, attribute: str | None = None, *, layout: LAYOUT = "csr", legacy: bool = False
    ) -> sps.sparray | sps.spmatrix:
        """
        Get this relationship matrix as a SciPy sparse matrix.

        .. note::
            If the selected attribute has missing values, they are *omitted* from the
            returned matrix.

        Args:
            attribute:
                The attribute to return, or ``None`` to return an indicator-only
                sparse matrix (all observed values are 1).
            layout:
                The matrix layout to return.

        Returns:
            The sparse matrix.
        """
        n_rows = self.n_rows
        n_cols = self.n_cols
        nnz = self._table.num_rows

        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        mask = None
        if attribute is None or (attribute == "count" and "count" not in self._table.column_names):
            values = np.ones(nnz, dtype=np.float32)
        else:
            value_col = self._table.column(attribute)
            if value_col.null_count:
                mask = value_col.is_valid()
                values = value_col.filter(mask).to_numpy()
                mask = mask.to_numpy()
            else:
                values = value_col.to_numpy()

        if layout == "csr" and mask is None:
            if legacy:
                return sps.csr_matrix((values, colinds, self._row_ptrs), shape=(n_rows, n_cols))
            else:
                return sps.csr_array((values, colinds, self._row_ptrs), shape=(n_rows, n_cols))
        else:
            rowinds = self._table.column(num_col_name(self.row_type)).to_numpy()
            if mask is not None:
                colinds = colinds[mask]
                rowinds = rowinds[mask]
            if legacy:
                mat = sps.coo_matrix((values, (rowinds, colinds)), shape=(n_rows, n_cols))
                if layout == "csr":
                    mat = mat.tocsr()
            else:
                mat = sps.coo_array((values, (rowinds, colinds)), shape=(n_rows, n_cols))
                if layout == "csr":
                    mat = mat.tocsr()
            return mat

    @overload
    def torch(
        self, attribute: str | None = None, *, layout: Literal["csr"] = "csr"
    ) -> torch.Tensor: ...
    @overload
    def torch(self, attribute: str | None = None, *, layout: Literal["coo"]) -> torch.Tensor: ...
    def torch(self, attribute: str | None = None, *, layout: LAYOUT = "csr") -> torch.Tensor:
        """
        Get this relationship matrix as a PyTorch sparse tensor.

        .. note::
            If the selected attribute has missing values, they are *omitted* from the
            returned matrix.

        Args:
            attribute:
                The attribute to return, or ``None`` to return an indicator-only
                sparse matrix (all observed values are 1).
            layout:
                The matrix layout to return.

        Returns:
            The sparse matrix.
        """
        n_rows = self.n_rows
        n_cols = self.n_cols
        nnz = self._table.num_rows

        colinds = self._table.column(num_col_name(self.col_type)).to_numpy()
        colinds = torch.tensor(np.require(colinds, requirements="W"))
        mask = None
        if attribute is None or (attribute == "count" and "count" not in self._table.column_names):
            values = torch.ones(nnz, dtype=torch.float32)
        else:
            mask = None
            value_col = self._table.column(attribute)
            if value_col.null_count:
                mask = value_col.is_valid()
                value_col = value_col.filter(mask)
                mask = mask.to_numpy()

            if pa.types.is_timestamp(self._table.field(attribute).type):
                value_col = value_col.cast(pa.timestamp("s")).cast(pa.int64())

            values = torch.tensor(value_col.to_numpy())

        if layout == "csr" and mask is None:
            return torch.sparse_csr_tensor(
                crow_indices=torch.tensor(self._row_ptrs),
                col_indices=colinds,
                values=values,
                size=(n_rows, n_cols),
            )
        else:
            rowinds = torch.tensor(self._table.column(num_col_name(self.row_type)).to_numpy())
            indices = torch.stack((rowinds, colinds))
            if mask is not None:
                indices = indices[:, torch.as_tensor(mask)]
            mat = torch.sparse_coo_tensor(
                indices=indices, values=values, size=(n_rows, n_cols)
            ).coalesce()
            if layout == "csr":
                mat = mat.to_sparse_csr()
            return mat

    def sample_negatives(
        self,
        rows: np.ndarray[tuple[int], np.dtype[np.int32]],
        *,
        weighting: Literal["uniform", "popular", "popularity"] = "uniform",
        n: int | None = None,
        verify: bool = True,
        max_attempts: int = 10,
        rng: np.random.Generator | None = None,
    ) -> NDArray[np.int32]:
        """
        Sample negative columns (columns with no observation recorded) for an
        array of rows. On a normal interaction matrix, this samples negative
        items for users.

        Args:
            rows:
                The row numbers.  Duplicates are allowed, and negative columns
                are sampled independently for each row. Must be a 1D array or
                tensor.
            weighting:
                The weighting for sampled negatives; ``uniform`` samples them
                uniformly at random, while ``popularity`` samples them
                proportional to their popularity (number of occurrences).
            n:
                The number of negatives to sample for each user.  If ``None``,
                a single-dimensional vector is returned.
            verify:
                Whether to verify that the negative items are actually negative.
                Unverified sampling is much faster but can return false
                negatives.
            max_attempts:
                When verification is on, the maximum attempts before giving up
                and returning a possible false negative.
            rng:
                A random number generator to use.
        """
        rng = random_generator(rng)

        _log.debug("samping negatives", nrows=len(rows), ncols=n)

        eff_n = n or 1

        if verify:
            rows = np.require(rows, np.int32)
            columns = _accel_data.sample_negatives(
                self._coords,
                rows,
                self.n_cols,
                n=eff_n,
                max_attempts=max_attempts,
                pop_weighted=weighting != "uniform",
                seed=rng.bit_generator.random_raw(),
            )

        elif weighting == "uniform":
            columns = self._sample_unweighted(rng, eff_n)
        elif weighting == "popular" or weighting == "popularity":
            columns = self._sample_weighted(rng, eff_n)
        else:  # pragma: nocover
            raise ValueError(f"unsupported weighting {weighting}")

        columns = np.require(columns, np.int32)
        if n is None:
            columns = columns.reshape(-1)
        else:
            columns = columns.reshape(-1, n)

        return columns

    def _sample_unweighted(self, rng: np.random.Generator, size: int | tuple[int, int]):
        return rng.integers(0, self.n_cols, size=size)

    def _sample_weighted(self, rng: np.random.Generator, size: int | tuple[int, int]):
        return rng.choice(self._col_nums.to_numpy(), size=size, replace=False)

    @overload
    def row_table(self, id: ID) -> pa.Table | None: ...
    @overload
    def row_table(self, *, number: int) -> pa.Table: ...
    def row_table(self, id: ID | None = None, *, number: int | None = None) -> pa.Table | None:
        """
        Get a single row of this interaction matrix as a table.
        """
        if number is None and id is None:  # pragma: noover
            raise ValueError("must provide one of id and number")

        if number is None:
            number = self.row_vocabulary.number(id, "none")
            if number is None:
                return None

        row_start = self._row_ptrs[number]
        row_end = self._row_ptrs[number + 1]

        tbl = self._table.slice(row_start, row_end - row_start)
        tbl = tbl.drop_columns(num_col_name(self.row_type))
        return tbl

    @overload
    def row_items(self, id: ID) -> ItemList | None: ...
    @overload
    def row_items(self, *, number: int) -> ItemList: ...
    def row_items(self, id: ID | None = None, *, number: int | None = None) -> ItemList | None:
        """
        Get a single row of this interaction matrix as an item list.  Only valid
        when the column entity class is ``item''.
        """
        if self.col_type != "item":
            raise RuntimeError("row_items() only valid for item-column matrices")

        tbl = self.row_table(id=id, number=number)  # type: ignore
        if tbl is None:
            return None

        return ItemList.from_arrow(tbl, vocabulary=self.col_vocabulary)

    def to_ilc(self) -> ItemListCollection:
        """
        Get the rows as an item list collection.
        """
        from .collection import ListILC

        if self.col_type != "item":
            raise RuntimeError("row_items() only valid for item-column matrices")

        # FIXME: make this a lot faster with Arrow
        ilc = ListILC([f"{self.row_type}_id"])
        for i in range(self.n_rows):
            ilc.add(self.row_items(number=i), self.row_vocabulary.id(i))
        return ilc

    def row_stats(self):
        if self._row_stats is None:
            self._row_stats = self._compute_stats(self.row_type, self.col_type, self.row_vocabulary)
        return self._row_stats

    def col_stats(self):
        if self._col_stats is None:
            self._col_stats = self._compute_stats(self.col_type, self.row_type, self.col_vocabulary)
        return self._col_stats

    def _compute_stats(
        self, stat_type: str, other_type: str, stat_vocab: Vocabulary
    ) -> pd.DataFrame:
        s_col = num_col_name(stat_type)
        group = self._table.group_by(s_col)
        o_col = num_col_name(other_type)
        aggs = [(o_col, "count"), (o_col, "count_distinct")]
        if "count" in self._table.column_names:
            aggs.append(("count", "sum"))
        if "rating" in self._table.column_names:
            aggs += [("rating", "count"), ("rating", "mean")]
        if "first_time" in self._table.column_names:
            aggs.append(("first_time", "min"))
        if "last_time" in self._table.column_names:
            aggs.append(("last_time", "max"))
        if "timestamp" in self._table.column_names:
            aggs += [("timestamp", "min"), ("timestamp", "max")]

        stats = group.aggregate(aggs).to_pandas()
        stats = stats.rename(
            columns={
                o_col + "_count": "record_count",
                o_col + "_count_distinct": other_type + "_count",
                "count_sum": "count",
                "rating_mean": "mean_rating",
                "first_time_min": "first_time",
                "last_time_max": "last_time",
            }
        )
        if "count" not in stats.columns:
            stats["count"] = stats["record_count"]
        if "timestamp" in self._table.column_names:
            if "first_time" in self._table.columns:
                stats["first_time"] = np.minimum(stats["first_time"], stats["timestamp_min"])
            else:
                stats["first_time"] = stats["timestamp_min"]
            del stats["timestamp_min"]

            if "last_time" in self._table.columns:
                stats["last_time"] = np.maximum(stats["last_time"], stats["timestamp_max"])
            else:
                stats["last_time"] = stats["timestamp_max"]
            del stats["timestamp_max"]

        id_col = id_col_name(stat_type)
        stats[id_col] = stat_vocab.ids(stats[s_col])
        del stats[s_col]
        stats.set_index(id_col, inplace=True)
        stats = stats.reindex(stat_vocab.index, fill_value=0)
        if "mean_rating" in stats.columns:
            stats.loc[stats["rating_count"] == 0, "mean_rating"] = np.nan
        if "first_time" in stats.columns:
            stats.loc[stats["count"] == 0, "first_time"] = pd.NaT
        if "last_time" in stats.columns:
            stats.loc[stats["count"] == 0, "last_time"] = pd.NaT

        return stats



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.stats import kendalltau, permutation_test

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given, settings

from lenskit.basic import PopScorer, SoftmaxRanker
from lenskit.data.items import ItemList
from lenskit.testing import BasicComponentTests, ScorerTests, scored_lists

_log = logging.getLogger(__name__)


class TestSoftmax(BasicComponentTests):
    component = SoftmaxRanker


@given(scored_lists())
def test_unlimited_ranking(items: ItemList):
    topn = SoftmaxRanker()
    ranked = topn(items=items)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = np.isnan(scores)
    _log.info("ranking %d items, %d invalid", len(ids), np.sum(invalid))

    assert isinstance(ranked, ItemList)
    assert len(ranked) <= len(items)
    assert ranked.ordered
    # all valid items are included
    assert len(ranked) == np.sum(~invalid)

    # the set of valid items matches
    assert set(ids[~invalid]) == set(ranked.ids())

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_configured_truncation(n, items: ItemList):
    topn = SoftmaxRanker(n=n)
    ranked = topn(items=items)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = np.isnan(scores)
    _log.info("top %d of %d items, %d invalid", n, len(ids), np.sum(invalid))

    val_items = items[~invalid]

    assert isinstance(ranked, ItemList)
    assert ranked.ordered
    assert len(ranked) == min(n, len(val_items))

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None
    src_s = src_s[src_s.notna()]

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_runtime_truncation(n, items: ItemList):
    topn = SoftmaxRanker(rng="user")
    ranked = topn(items=items, n=n)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = np.isnan(scores)
    _log.info("top %d of %d items, %d invalid", n, len(ids), np.sum(invalid))

    val_items = items[~invalid]

    assert isinstance(ranked, ItemList)
    assert ranked.ordered
    assert len(ranked) == min(n, len(val_items))

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None
    src_s = src_s[src_s.notna()]

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)


def test_stochasticity(rng):
    "Test that softmax is varying but order-consistent"
    iids = np.arange(500)
    scores = rng.normal(size=500)
    scores = np.square(scores)
    items = ItemList(item_ids=iids, scores=scores)
    size = 50

    TRIALS = 100
    topn = SoftmaxRanker(n=size)

    _log.info("testing stochastic ranking: top %d of %d", size, len(items))

    ranks = np.full((size, TRIALS), -1, dtype=np.int64)
    scores = np.full((size, TRIALS), np.nan, dtype=np.float64)
    for i in range(TRIALS):
        ranked = topn(items)
        assert len(ranked) == size
        ranks[:, i] = ranked.ids()
        scores[:, i] = ranked.scores()

    id_counts = np.array([len(np.unique(ranks[i, :])) for i in range(size)])
    try:
        # at least half the positions should have more than 5 different items show up
        assert np.sum(id_counts < 5) <= size / 2
    except AssertionError as e:
        _log.info("failed test with n=%d, N=%d", size, len(items))
        _log.info("item counts: %s", id_counts)
        _log.info("items:\n%s", items.to_df())
        raise e

    # We want to test that it is usually putting things in the correct order.
    # We'll do this by computing Kendall's tau between the sampled ranking and
    # the scores of those items.  We want most of the lists (90%) to have
    # correlations statistically significnatly greater than zero (items tend to
    # be in the correct order).
    pvals = []
    taus = []
    for i in range(TRIALS):
        r_items = ranks[:, i]
        ranked = r_items >= 0
        r_items = r_items[ranked]

        rii = items[r_items]
        trs = size - np.arange(size)

        tau = kendalltau(trs, rii.scores(), alternative="greater")
        taus.append(tau.statistic)
        pvals.append(tau.pvalue)
        _log.info("trial %d: 𝜏=%.3f, p=%.3f", i, tau.statistic, tau.pvalue)

    pvals = np.array(pvals)
    _log.info("trial p-value statistics: mean=%.3f, median=%.3f", np.mean(pvals), np.median(pvals))
    # do 90% of trials pass the test?
    # ALERT: this is broken
    assert np.sum(pvals < 0.05) >= 0.9



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit ranking (and list) metrics.
"""

from ._base import RankingMetricBase
from ._dcg import DCG, NDCG
from ._gini import ExposureGini, ListGini
from ._hit import Hit
from ._map import AveragePrecision
from ._pop import MeanPopRank
from ._pr import Precision, Recall
from ._rbp import RBP
from ._recip import RecipRank
from ._weighting import GeometricRankWeight, LogRankWeight, RankWeight

__all__ = [
    "RankingMetricBase",
    "RankWeight",
    "GeometricRankWeight",
    "LogRankWeight",
    "Hit",
    "Precision",
    "Recall",
    "RecipRank",
    "NDCG",
    "DCG",
    "RBP",
    "MeanPopRank",
    "AveragePrecision",
    "ListGini",
    "ExposureGini",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from uuid import UUID, uuid4

if TYPE_CHECKING:
    from ..multiprocess._protocol import ProgressMessage


class Progress:
    """
    Base class for progress reporting.  The default implementations do nothing.
    """

    uuid: UUID
    total: int | float | None

    def __init__(self, *args: Any, uuid: UUID | None = None, **kwargs: Any):
        self.uuid = uuid if uuid is not None else uuid4()

    @classmethod
    def handle_message(cls, update: ProgressMessage):
        pass

    def update(
        self,
        advance: int = 1,
        completed: int | None = None,
        total: int | None = None,
        **kwargs: float | int | str,
    ):
        """
        Update the progress bar.

        Args:
            advance:
                The amount to advance by.
            completed:
                The number completed; this overrides ``advance``.
            total:
                A new total, to update the progress bar total.
        """
        pass

    def finish(self):
        """
        Finish and clean up this progress bar.  If the progresss bar is used as
        a context manager, this is automatically called on context exit.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args: Any):
        self.finish()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from typing import Iterator, overload

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from lenskit.data import NPID, Dataset, DatasetBuilder, ItemListCollection, UserIDKey
from lenskit.logging import item_progress
from lenskit.random import RNGInput, random_generator

from .holdout import HoldoutMethod
from .split import TTSplit

_log = logging.getLogger(__name__)


def crossfold_users(
    data: Dataset,
    partitions: int,
    method: HoldoutMethod,
    *,
    test_only: bool = False,
    rng: RNGInput | None = None,
) -> Iterator[TTSplit]:
    """
    Partition a frame of ratings or other data into train-test partitions
    user-by-user. This function does not care what kind of data is in `data`, so
    long as it is a Pandas DataFrame (or equivalent) and has a `user` column.

    Stability:
        Caller

    Args:
        data:
            a data frame containing ratings or other data you wish to partition.
        partitions:
            the number of partitions to produce
        method:
            The method for selecting test rows for each user.
        test_only:
            If ``True``, returns splits with only testing data.
        rng:
            The random number generator or seed (see :ref:`rng`).

    Returns
        The train-test pairs.
    """
    rng = random_generator(rng)

    users = data.users.ids()
    _log.info(
        "partitioning %d rows for %d users into %d partitions",
        data.interactions().count(),
        len(users),
        partitions,
    )

    # create an array of indexes into user row
    rows = np.arange(len(users))
    # shuffle the indices & split into partitions
    rng.shuffle(rows)
    test_sets = np.array_split(rows, partitions)

    # get the whole test DF
    df = data.interaction_matrix(format="pandas", original_ids=True)

    # convert each partition into a split
    for i, ts in enumerate(test_sets):
        # get our users!
        test_us = users[ts]
        _log.info("fold %d: selecting test ratings", i)

        yield _make_split(data, df, test_us, method, test_only=test_only)


@overload
def sample_users(
    data: Dataset,
    size: int,
    method: HoldoutMethod,
    *,
    repeats: int,
    disjoint: bool = True,
    test_only: bool = False,
    rng: RNGInput = None,
) -> Iterator[TTSplit]: ...
@overload
def sample_users(
    data: Dataset,
    size: int,
    method: HoldoutMethod,
    *,
    disjoint: bool = True,
    rng: RNGInput = None,
    test_only: bool = False,
    repeats: None = None,
) -> TTSplit: ...
def sample_users(
    data: Dataset,
    size: int,
    method: HoldoutMethod,
    *,
    repeats: int | None = None,
    disjoint: bool = True,
    test_only: bool = False,
    rng: RNGInput = None,
) -> Iterator[TTSplit] | TTSplit:
    """
    Create train-test splits by sampling users.  When ``repeats`` is None,
    returns a single train-test split; otherwise, it returns an iterator over
    multiple splits. If ``repeats=1``, this function returns an iterator that
    yields a single train-test pair.

    Stability:
        Caller

    Args:
        data:
            Data frame containing ratings or other data you wish to partition.
        size:
            The sample size.
        method:
            The method for obtaining user test ratings.
        repeats:
            The number of samples to produce.
        test_only:
            If ``True``, returns splits with empty training sets (useful when
            you just want to save the test data).
        rng:
            The random number generator or seed (see :ref:`rng`).

    Returns:
        The train-test pair(s).
    """

    rng = random_generator(rng)

    users = data.users.ids()
    unums = np.arange(len(users))
    if disjoint and repeats is not None and repeats * size >= len(users):
        _log.warning(
            "cannot take %d disjoint samples of size %d from %d users", repeats, size, len(users)
        )
        return crossfold_users(data, repeats, method)

    _log.info("sampling %d users (n=%d)", len(users), size)

    # get the whole test DF
    rate_df = data.interactions().pandas(ids=True)

    if repeats is None:
        test_us = rng.choice(users, size, replace=False)
        return _make_split(data, rate_df, test_us, method)

    if disjoint:
        rng.shuffle(unums)
        test_usets = [unums[i * size : (i + 1) * size] for i in range(repeats)]
    else:
        test_usets = [rng.choice(len(users), size, replace=False) for _i in range(repeats)]

    return (_make_split(data, rate_df, users[us], method, test_only=test_only) for us in test_usets)


def _make_split(
    data: Dataset,
    df: pd.DataFrame,
    test_us: NDArray[NPID],
    method: HoldoutMethod,
    *,
    test_only: bool = False,
) -> TTSplit:
    # create the test sets for these users
    test = ItemListCollection(UserIDKey)

    with item_progress("Splitting users", len(test_us)) as pb:
        for u in test_us:
            row = data.user_row(u)
            assert row is not None
            u_test = method(row)
            test.add(u_test, u)
            pb.update()

    train_build = DatasetBuilder(data)
    iname = data.default_interaction_class()
    if test_only:
        train_build.clear_relationships(iname)
    else:
        test_tbl = test.to_df()[["user_id", "item_id"]]
        train_build.filter_interactions(iname, remove=test_tbl)

    return TTSplit(train_build.build(), test)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
from os import fspath
from pathlib import Path

from click.testing import CliRunner

from pytest import mark

from lenskit.cli import lenskit
from lenskit.data.adapt import from_interactions_df
from lenskit.pipeline.config import PipelineConfig


@mark.slow
@mark.realdata
def test_tune_cli(tmpdir: Path, ml_100k):
    tmpdir = Path(tmpdir)
    ml_ds = from_interactions_df(ml_100k)
    ds_path = tmpdir / "ml-data"
    ml_ds.save(ds_path)

    train_file = tmpdir / "ml-data.train"
    test_file = tmpdir / "ml-data.test.parquet"
    out_dir = tmpdir / "bias-tune"

    runner = CliRunner()
    result = runner.invoke(
        lenskit, ["data", "split", "--fraction=0.2", "--min-train-interactions=5", fspath(ds_path)]
    )
    assert result.exit_code == 0

    result = runner.invoke(
        lenskit,
        [
            "tune",
            "-T",
            fspath(train_file),
            "-V",
            fspath(test_file),
            "--save-pipeline",
            fspath(tmpdir / "pipeline.json"),
            "pipelines/bias-search.toml",
            fspath(out_dir),
        ],
    )

    assert result.exit_code == 0
    assert out_dir.exists()
    assert (out_dir / "result.json").exists()
    assert (tmpdir / "pipeline.json").exists()

    cfg = PipelineConfig.model_validate_json((tmpdir / "pipeline.json").read_text())
    res = json.loads((out_dir / "result.json").read_text())

    s_cfg = cfg.components["scorer"].config
    assert s_cfg is not None
    assert s_cfg["damping"] == res["config"]["damping"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic data types used in data representations.
"""

# pyright: strict
from __future__ import annotations

from dataclasses import dataclass, field
from typing import NamedTuple

import numpy as np
import pandas as pd
import pyarrow as pa
from typing_extensions import Any, Generic, Literal, Sequence, TypeAlias, TypeVar

FeedbackType: TypeAlias = Literal["explicit", "implicit"]
"Types of feedback supported."

CoreID: TypeAlias = int | str | bytes
"Core (non-NumPy) identifier types."
NPID: TypeAlias = np.integer[Any] | np.str_ | np.bytes_ | np.object_
"NumPy entity identifier types."
ID: TypeAlias = CoreID | NPID
"Allowable identifier types."
IDArray: TypeAlias = np.ndarray[tuple[int], np.dtype[NPID]]
"NumPy arrays of identifiers."
IDSequence: TypeAlias = (
    Sequence[ID]
    | IDArray
    | pa.StringArray
    | "pa.IntegerArray[Any]"
    | "pa.ChunkedArray[Any]"
    | "pd.Series[CoreID]"
)
"Sequences of identifiers."

V = TypeVar("V", bound=np.number[Any], default=np.float32)
NPMatrix: TypeAlias = np.ndarray[tuple[int, int], np.dtype[V]]
NPVector: TypeAlias = np.ndarray[tuple[int], np.dtype[V]]

T = TypeVar("T")

DF_FORMAT: TypeAlias = Literal["numpy", "pandas", "torch"]
MAT_FORMAT: TypeAlias = Literal["scipy", "torch", "pandas", "structure"]
MAT_AGG: TypeAlias = Literal["count", "sum", "mean", "first", "last"]
LAYOUT: TypeAlias = Literal["csr", "coo"]


@dataclass
class AliasedColumn:
    """
    A data frame column with possible aliases.

    Stability:
        Testing
    """

    name: str
    "The column name."
    compat_aliases: list[str] = field(default_factory=list)
    "A list of aliases for the column."
    warn: bool = False
    "Whether to warn when using an alias."


Column: TypeAlias = str | AliasedColumn


@dataclass(frozen=True)
class UIPair(Generic[T]):
    """
    A user-item pair of values.
    """

    user: T
    item: T


class Extent(NamedTuple):
    """
    Representation of a range with a start and end.
    """

    start: int
    "The range start (inclusive)."
    end: int
    "The range end (exclusive)."

    @property
    def size(self) -> int:
        """
        The size of the extent.
        """
        return self.end - self.start



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit test harnesses and utilities.

This package contains utility code for testing LensKit and its components,
including in derived packages.  It relies on PyTest and Hypothesis.
"""

import os
from contextlib import contextmanager

import numpy as np

import hypothesis.strategies as st

from ._arrays import coo_arrays, scored_lists, sparse_arrays, sparse_tensors
from ._components import BasicComponentTests, ScorerTests
from ._movielens import (
    DemoRecs,
    demo_recs,
    ml_20m,
    ml_100k,
    ml_100k_zip,
    ml_ds,
    ml_ds_unchecked,
    ml_ratings,
    ml_test_dir,
    pop_recs,
)

__all__ = [
    "coo_arrays",
    "scored_lists",
    "sparse_arrays",
    "sparse_tensors",
    "ml_100k",
    "ml_100k_zip",
    "ml_20m",
    "ml_ds",
    "ml_ds_unchecked",
    "ml_ratings",
    "ml_test_dir",
    "demo_recs",
    "pop_recs",
    "set_env_var",
    "DemoRecs",
    "BasicComponentTests",
    "ScorerTests",
]


@contextmanager
def set_env_var(var, val):
    "Set an environment variable & restore it."
    old_val = os.environ.get(var, None)
    try:
        if val is None:
            if old_val is not None:
                del os.environ[var]
        else:
            os.environ[var] = val
        yield
    finally:
        if old_val is not None:
            os.environ[var] = old_val
        elif val is not None:
            del os.environ[var]


def have_memory(gb: int | float) -> bool:
    "Check if we have at least gb gigs of memory."
    if not hasattr(os, "sysconf"):
        return False

    p_size = os.sysconf("SC_PAGE_SIZE")
    p_count = os.sysconf("SC_PHYS_PAGES")
    mem_size = p_size * p_count
    return mem_size >= gb * 1024 * 1024 * 1024


def integer_ids():
    """
    Hypothesis strategy to generate valid integer user/item IDs.
    """
    return st.integers(1, np.iinfo("i8").max)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Graph-based models, especially GNNs with :mod:`torch_geometric`.
"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules="lightgcn",
    submod_attrs={
        "lightgcn": ["LightGCNScorer", "LightGCNConfig"],
    },
)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import datetime as dt

import numpy as np

from pytest import approx

from lenskit.data import Dataset
from lenskit.splitting import split_global_time, split_temporal_fraction


def test_temporal_split(ml_ds: Dataset):
    point = dt.datetime.fromisoformat("2015-01-01")

    split = split_global_time(ml_ds, "2015-01-01")

    n_test = sum(len(il) for il in split.test.lists())
    assert n_test + split.train.interaction_count == ml_ds.interaction_count

    assert np.all(split.train.interaction_table(format="pandas")["timestamp"] < point)
    for u, il in split.test:
        ts = il.field("timestamp", format="pandas")
        assert ts is not None
        assert np.all(ts >= point)


def test_temporal_split_limit(ml_ds: Dataset):
    split = split_global_time(ml_ds, "2015-01-01", filter_test_users=True)

    tr_count = split.train.user_stats()

    for k in split.test.keys():
        (u,) = k
        assert tr_count.loc[u, "rating_count"] >= 1


def test_temporal_split_limit2(ml_ds: Dataset):
    split = split_global_time(ml_ds, "2015-01-01", filter_test_users=2)

    tr_count = split.train.user_stats()

    for k in split.test.keys():
        (u,) = k
        assert tr_count.loc[u, "rating_count"] >= 2


def test_temporal_split_fraction(ml_ds: Dataset):
    split = split_temporal_fraction(ml_ds, 0.2)

    n_test = sum(len(il) for il in split.test.lists())
    assert n_test + split.train.interaction_count == ml_ds.interaction_count
    assert n_test / ml_ds.interaction_count == approx(0.2, abs=0.01)

    point = min(il.field("timestamp").min() for il in split.test.lists())
    assert np.all(split.train.interaction_table(format="pandas")["timestamp"] < point)


def test_multi_split(ml_ds: Dataset):
    p1 = dt.datetime.fromisoformat("2015-01-01")
    p2 = dt.datetime.fromisoformat("2016-01-01")

    valid, test = split_global_time(ml_ds, ["2015-01-01", "2016-01-01"])

    n_test = sum(len(il) for il in test.test.lists())
    n_valid = sum(len(il) for il in valid.test.lists())
    assert n_test + test.train.interaction_count == ml_ds.interaction_count
    assert n_test + n_valid + valid.train.interaction_count == ml_ds.interaction_count

    assert np.all(valid.train.interaction_table(format="pandas")["timestamp"] < p1)
    assert np.all(test.train.interaction_table(format="pandas")["timestamp"] < p2)
    for u, il in test.test:
        ts = il.field("timestamp", format="pandas")
        assert ts is not None
        assert np.all(ts >= p2)

    for u, il in valid.test:
        ts = il.field("timestamp", format="pandas")
        assert ts is not None
        assert np.all(ts >= p1)
        assert np.all(ts < p2)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import pandas as pd

from lenskit.pipeline import PipelineBuilder, PipelineProfiler

from .test_pipeline_clone import PrefixConfig, Prefixer


def test_basic_profile(tmpdir: Path):
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.add_component("prefix", Prefixer, PrefixConfig(prefix="scroll named "), msg=msg)
    builder.default_component("prefix")
    pipe = builder.build()

    pipe_file = tmpdir / "profile.csv"
    with PipelineProfiler(pipe, pipe_file) as profiler:
        pipe.run("prefix", msg="hello", _profile=profiler)

    assert pipe_file.exists()
    print(pipe_file.read_text("utf8"))
    df = pd.read_csv(pipe_file)
    assert df.columns == ["prefix"]
    assert len(df) == 1

    # very short time - less than 100ms
    assert df.loc[0, "prefix"] < 100000



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Components that look up user history from the training data.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import numpy as np
import pandas as pd
from typing_extensions import override

from lenskit.data import Dataset, ItemList, QueryInput, RecQuery
from lenskit.data.dataset import MatrixRelationshipSet
from lenskit.diagnostics import DataError
from lenskit.logging import get_logger, trace
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


@dataclass
class LookupConfig:
    interaction_class: str | None = None
    """
    The name of the interaction class to use.  Leave ``None`` to use the
    dataset's default interaction class.
    """


class UserTrainingHistoryLookup(Component[ItemList], Trainable):
    """
    Look up a user's history from the training data.

    Stability:
        Caller
    """

    config: LookupConfig

    interactions: MatrixRelationshipSet

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        # TODO: find a better data structure for this
        if hasattr(self, "interactions") and not options.retrain:
            return

        self.interactions = data.interactions(self.config.interaction_class).matrix()
        if self.interactions.row_type != "user":  # pragma: nocover
            raise DataError("interactions must have user rows")
        if self.interactions.col_type != "item":  # pragma: nocover
            raise DataError("interactions must have item columns")

    def __call__(self, query: QueryInput) -> RecQuery:
        """
        Look up the user's data from the training history (if needed), and
        ensure a fully-populated :class:`RecQuery`.
        """
        query = RecQuery.create(query)
        if query.user_id is None:
            trace(_log, "no user ID")
            return query

        log = _log.bind(user_id=query.user_id)
        id_type = self.interactions.row_vocabulary.ids().dtype
        if isinstance(query.user_id, str) and issubclass(id_type.type, np.number):
            query.user_id = id_type.type(query.user_id)

        if query.user_items is None:
            trace(log, "looking up user history")
            query.user_items = self.interactions.row_items(query.user_id)
            if query.user_items is not None:
                log.debug("fetched %d history items", len(query.user_items))
            else:
                log.debug("no history found")

        return query


@dataclass
class KnownRatingConfig(LookupConfig):
    score: Literal["rating", "indicator"] | None = None
    """
    The field name to use to score items, or ``"indicator"`` to score with 0/1
    based on presence in the training data.  The default, ``None``, uses ratings
    if available, and otherwise scores with ` for interacted items and leaves
    un-interacted items unscored.
    """
    source: Literal["training", "query"] = "training"
    """
    Whether to get the known ratings from the training data or from the query.
    """


class KnownRatingScorer(Component[ItemList], Trainable):
    """
    Score items by returning their values from the training data.

    Stability:
        Caller
    """

    config: KnownRatingConfig
    interactions: MatrixRelationshipSet

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "interactions") and not options.retrain:
            return

        if self.config.source == "query":
            return

        self.interactions = data.interactions(self.config.interaction_class).matrix()

    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        # figure out what scores we start with
        ilist = None
        if self.config.source == "query" and query.user_items is not None:
            ilist = query.user_items

        elif (
            self.config.source == "training"
            and query.user_id is not None
            and query.user_id in self.interactions.row_vocabulary
        ):
            ilist = self.interactions.row_items(query.user_id)

        if ilist is None:
            scores = None
        elif self.config.score == "indicator":
            scores = pd.Series(1.0, index=ilist.ids())
        else:
            scores = ilist.field("rating", format="pandas", index="ids")
            if scores is None and self.config.score is None:
                scores = pd.Series(1.0, index=ilist.ids())

        if scores is None:
            scores = pd.Series(np.nan, index=items.ids())

        scores = scores.reindex(
            items.ids(), fill_value=0.0 if self.config.score == "indicator" else np.nan
        )
        return ItemList(items, scores=scores.values)  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd
from numpy.random import Generator

from lenskit.data import Dataset


def test_all_entities(rng: Generator, ml_ratings: pd.DataFrame, ml_ds: Dataset):
    assert len(ml_ds.entities("item")) == ml_ds.item_count
    assert len(ml_ds.entities("item")) >= ml_ratings["item_id"].nunique()
    assert len(ml_ds.entities("user")) == ml_ratings["user_id"].nunique()

    stats = ml_ds.item_stats()

    assert np.all(
        ml_ds.entities("item").ids()[stats["count"] > 0] == np.unique(ml_ratings["item_id"])
    )
    assert np.all(ml_ds.entities("item").numbers() == np.arange(ml_ds.item_count))

    df = ml_ds.entities("item").pandas()
    assert len(df) >= ml_ratings["item_id"].nunique()
    assert np.all(df["item_id"] == ml_ds.items.ids())


def test_entity_subset_ids(rng: Generator, ml_ratings: pd.DataFrame, ml_ds: Dataset):
    item_ids = rng.choice(ml_ratings["item_id"].unique(), 20, replace=False)

    ents = ml_ds.entities("item").select(ids=item_ids)
    assert len(ents) == len(item_ids)
    assert np.all(ents.ids() == item_ids)
    assert np.all(ents.numbers() == ml_ds.items.numbers(item_ids))

    df = ents.pandas()
    assert np.all(df["item_id"] == item_ids)


def test_entity_subset_numbers(rng: Generator, ml_ratings: pd.DataFrame, ml_ds: Dataset):
    inos = rng.choice(ml_ratings["item_id"].nunique(), 20, replace=False)

    ents = ml_ds.entities("item").select(numbers=inos)
    assert len(ents) == len(inos)
    assert np.all(ents.numbers() == inos)
    assert np.all(ents.ids() == ml_ds.items.ids(inos))


def test_entity_subset_subset_numbers(rng: Generator, ml_ratings: pd.DataFrame, ml_ds: Dataset):
    "Test that subsetting a subset works."
    inos = rng.choice(ml_ratings["item_id"].nunique(), 100, replace=False)

    ents = ml_ds.entities("item").select(numbers=inos)
    assert len(ents) == len(inos)

    iss2 = rng.choice(inos, 20, replace=False)
    e2 = ents.select(numbers=iss2)
    assert len(e2) == 20
    assert np.all(e2.ids() == ml_ds.items.ids(iss2))
    assert np.all(e2.numbers() == iss2)


def test_entity_drop_null_attributes(rng: Generator, ml_ds: Dataset):
    attr = ml_ds.entities("item").attribute("tag_counts")
    a2 = attr.drop_null()

    assert len(a2) < len(attr)

    mat = a2.scipy()
    assert np.all(np.diff(mat.indptr) >= 1)
    assert mat.data.sum() > 1000



#!/usr/bin/env python3
# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Update headers in source files.

Usage:
    update-headers.py [--check-only] [--error-on-change] [-y YEAR] [FILE...]

Options:
    -y YEAR, --year=YEAR
        Set the copyright year.
    --check-only
        Check headers, but don't update them.
    --error-on-change
        Exit with an error when headers are changed.
    FILE
        File to check or update.
"""

import re
import subprocess as sp
import sys
from datetime import date
from pathlib import Path

from docopt import docopt
from unbeheader.headers import SUPPORTED_FILE_TYPES, update_header
from unbeheader.typing import CommentSkeleton, SupportedFileType

SUPPORTED_FILE_TYPES["rs"] = SupportedFileType(
    re.compile(r"((^//|[\r\n]//).*)*"),
    CommentSkeleton("//", "//"),
)

options = docopt(__doc__ or "")
year = options["--year"]
files = options["FILE"]

if year is None:
    today = date.today()
    year = today.year

if files:
    files = [Path(p) for p in files]
else:
    out = sp.check_output(["git", "ls-files", "*.py", "*.rs"])

    files = [Path(p.strip()) for p in re.split(r"\r?\n", out.decode()) if p.strip()]

n = 0
print("scanning", len(files), "files")
for file in files:
    if update_header(file, year, check=options["--check-only"]):
        n += 1

print("updated", n, "files")
if options["--error-on-change"] and n > 0:
    sys.exit(5)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import pandas as pd

from pytest import mark

from lenskit.basic import BiasScorer
from lenskit.batch import BatchPipelineRunner
from lenskit.pipeline import PipelineProfiler, topn_pipeline
from lenskit.splitting.split import TTSplit

from .test_batch_pipeline import MLB, ml_split, mlb


@mark.parametrize(("ncpus"), [1, 2])
@mark.eval
def test_bias_batch(tmpdir: Path, ml_split: TTSplit, ncpus: int | None):
    algo = BiasScorer(damping=5)
    pipeline = topn_pipeline(algo, predicts_ratings=True, n=20)
    pipeline.train(ml_split.train)

    file = tmpdir / "profile.csv"

    with PipelineProfiler(pipeline, file) as profiler:
        runner = BatchPipelineRunner(n_jobs=ncpus, profiler=profiler)
        runner.recommend()

        _results = runner.run(pipeline, ml_split.test)

    prof = pd.read_csv(file)
    print(prof)
    assert list(prof.columns) == pipeline.component_names()
    assert len(prof) == len(ml_split.test)
    # is the scorer the right time? >1µs, <25ms
    assert prof["scorer"].mean() > 1000
    assert prof["scorer"].mean() < 25_000_000



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from __future__ import annotations

import json
import logging
import os
import warnings
from typing import Optional

import torch
from threadpoolctl import threadpool_limits

from lenskit._accel import init_accel_pool

_config: Optional[ParallelConfig] = None
_log = logging.getLogger(__name__)


class ParallelConfig:
    """
    Parallel processing configuration.
    """

    _processes: list[int]
    _threads: list[int]
    _backend_threads: list[int]

    def __init__(
        self, processes: int | list[int], threads: int | list[int], backend_threads: int | list[int]
    ):
        self._processes = _ensure_numlist(processes)
        self._threads = _ensure_numlist(threads)
        self._backend_threads = _ensure_numlist(backend_threads)

    @property
    def processes(self) -> int:
        """
        Get the number of workers this process may spawn.
        """
        return self._processes[0]

    @property
    def threads(self) -> int:
        """
        Get the number of direct threads for this worker.
        """
        return self._threads[0]

    @property
    def backend_threads(self) -> int:
        """
        Get the number of backend threads for this worker.
        """
        return self._backend_threads[0]

    @property
    def total_threads(self) -> int:
        """
        Get the total number of threads for this worker (training threads times
        backend threads).
        """
        return self.threads * self.backend_threads

    def env_vars(self) -> dict[str, str]:
        """
        Get this parallel configuration as a set of environment variables.  The
        set also includes ``OMP_NUM_THREADS`` to configure OMP early.
        """
        evs = {
            "LK_NUM_PROCS": str(self.processes),
            "LK_NUM_THREADS": str(self.threads),
            "LK_NUM_BACKEND_THREADS": str(self.backend_threads),
        }
        if "OMP_NUM_THREADS" not in os.environ:
            evs["OMP_NUM_THREADS"] = str(self.backend_threads)
        if "OPENBLAS_NUM_THREADS" not in os.environ:
            evs["OPENBLAS_NUM_THREADS"] = str(self.backend_threads)
        if "MKL_NUM_THREADS" not in os.environ:
            evs["MKL_NUM_THREADS"] = str(self.backend_threads)
        return evs

    def subprocess(
        self,
        processes: int | list[int] | None = None,
        threads: int | list[int] | None = None,
        backend_threads: int | list[int] | None = None,
    ) -> ParallelConfig:
        """
        Get the parallel configuration for a subprocess.
        """
        ncpus = effective_cpu_count()

        if processes is None:
            if len(self._processes) > 1:
                processes = self._processes[1:]
                np = processes[0]
            else:
                np = processes = 1

        if threads is None:
            if len(self._threads) > 1:
                threads = self._threads[1:]
            else:
                allowed = min(ncpus // np, 4)
                threads = min(allowed, self.threads)

        if backend_threads is None:
            if len(self._backend_threads) > 1:
                backend_threads = self._backend_threads[1:]
            else:
                allowed = min(ncpus // np, 4)
                backend_threads = min(allowed, self.backend_threads)

        return ParallelConfig(processes=processes, threads=threads, backend_threads=backend_threads)

    def __str__(self):
        obj = json.dumps(
            {
                "processes": self.processes,
                "threads": self.threads,
                "backend_threads": self.backend_threads,
            },
            indent=2,
        )
        return f"ParallelConfig {obj}"


def initialize(
    config: ParallelConfig | None = None,
    *,
    processes: int | list[int] | None = None,
    threads: int | list[int] | None = None,
    backend_threads: int | list[int] | None = None,
    child_threads: int | None = None,
):
    """
    Set up and configure LensKit parallelism.  This only needs to be called if
    you want to control when and how parallelism is set up; components using
    parallelism will call :func:`ensure_init`, which will call this function
    with its default arguments if it has not been called.

    .. seealso:: :ref:`parallel-config`

    Args:
        processes:
            The number of processes to use for multiprocessing evaluations (see
            :envvar:`LK_NUM_PROCS`)
        threads:
            The number of threads to use for parallel model training and similar
            operations (see :envvar:`LK_NUM_THREADS`).
        backend_threads:
            The number of threads underlying computational engines should use
            (see :envvar:`LK_NUM_BACKEND_THREADS`).
        child_threads:
            The number of threads backends are allowed to use in the worker
            processes in multiprocessing operations (see
            :envvar:`LK_NUM_CHILD_THREADS`). Deprecated.
    """
    global _config
    if _config:
        _log.warning("parallelism already initialized")
        return

    # our parallel computation doesn't work with FD sharing
    torch.multiprocessing.set_sharing_strategy("file_system")

    if config is None:
        _config = _resolve_parallel_config(processes, threads, backend_threads, child_threads)
    else:
        _config = config
    _log.debug("configuring for parallelism: %s", _config)

    if "OPENBLAS_NUM_THREADS" not in os.environ and "MKL_NUM_THREADS" not in os.environ:
        threadpool_limits(_config.backend_threads, "blas")
    try:
        torch.set_num_interop_threads(_config.threads)
    except RuntimeError as e:
        _log.warning("failed to configure Pytorch interop threads: %s", e)
        warnings.warn("failed to set interop threads", RuntimeWarning)

    try:
        init_accel_pool(_config.threads)
    except RuntimeError as e:
        _log.warning("failed to initialize Rayon backend: %s", e)

    try:
        torch.set_num_threads(_config.backend_threads)
    except RuntimeError as e:
        _log.warning("failed to configure Pytorch intra-op threads: %s", e)
        warnings.warn("failed to set intra-op threads", RuntimeWarning)


def ensure_parallel_init():
    """
    Make sure LensKit parallelism is configured, and configure with defaults if
    it is not.

    Components using parallelism or intensive computations should call this
    function before they begin training.
    """
    if not _config:
        initialize()


def get_parallel_config() -> ParallelConfig:
    """
    Ensure that parallelism is configured and return the configuration.
    """
    ensure_parallel_init()
    assert _config is not None
    return _config


def subprocess_config(
    processes: int | list[int] | None = None,
    threads: int | list[int] | None = None,
    backend_threads: int | list[int] | None = None,
) -> ParallelConfig:
    """
    Get a parallel configuration for a subprocess.
    """
    cfg = get_parallel_config()
    return cfg.subprocess(processes, threads, backend_threads)


def effective_cpu_count() -> int:
    """
    Return the effective CPU count using the best available data.  Tries the following in order:

    1.  :func:`os.process_cpu_count`
    2.  :func:`os.sched_getaffinity`
    3.  :func:`os.cpu_count`
    """

    if hasattr(os, "process_cpu_count"):
        return os.process_cpu_count()  # type: ignore
    elif hasattr(os, "sched_getaffinity"):
        return len(os.sched_getaffinity(0))  # type: ignore
    else:
        ncpus = os.cpu_count()
        if ncpus is not None:
            return ncpus
        else:
            _log.warning("no CPU count available, assumping single CPU")
            return 1


def _resolve_parallel_config(
    processes: int | list[int] | None = None,
    threads: int | list[int] | None = None,
    backend_threads: int | list[int] | None = None,
    child_threads: int | None = None,
) -> ParallelConfig:
    nprocs = os.environ.get("LK_NUM_PROCS", None)
    nthreads = os.environ.get("LK_NUM_THREADS", None)
    nbthreads = os.environ.get("LK_NUM_BACKEND_THREADS", None)
    cthreads = os.environ.get("LK_NUM_CHILD_THREADS", None)
    ncpus = effective_cpu_count()

    if processes is None and nprocs:
        processes = _parse_numlist(nprocs)

    if threads is None and nthreads:
        threads = _parse_numlist(nthreads)

    if backend_threads is None and nbthreads:
        backend_threads = _parse_numlist(nbthreads)

    if child_threads is None and cthreads:
        warnings.warn(
            "LK_NUM_CHILD_THREADS is deprecated, use list in LK_NUM_BACKEND_THREADS instead",
            DeprecationWarning,
        )
        child_threads = int(cthreads)
    elif child_threads is not None:
        warnings.warn(
            "child_theads is deprecated, use list in backend_threads instead",
            DeprecationWarning,
        )

    if processes is None:
        processes = min(ncpus, 4)

    if threads is None:
        threads = min(ncpus, 8)

    processes = _ensure_numlist(processes)
    threads = _ensure_numlist(threads)

    if backend_threads is None:
        backend_threads = max(min(ncpus // threads[0], 4), 1)

    backend_threads = _ensure_numlist(backend_threads)

    if child_threads is not None:
        backend_threads.append(child_threads)

    return ParallelConfig(processes, threads, backend_threads)


def _unparse_numlist(nums: list[int]) -> str:
    return ",".join([str(i) for i in nums])


def _parse_numlist(spec: str):
    return [int(s.strip()) for s in spec.split(",")]


def _ensure_numlist(nums: int | list[int] | str) -> list[int]:
    if isinstance(nums, str):
        return _parse_numlist(nums)
    elif isinstance(nums, list):
        return nums
    else:
        return [nums]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from lenskit.basic.candidates import (
    AllTrainingItemsCandidateSelector,
    UnratedTrainingItemsCandidateSelector,
)
from lenskit.data import Dataset, from_interactions_df
from lenskit.testing import ml_ds, ml_ratings  # noqa: F401


def test_all(ml_ds: Dataset):
    sel = AllTrainingItemsCandidateSelector()
    sel.train(ml_ds)

    cands = sel()
    assert len(cands) == ml_ds.item_count
    assert np.all(cands.ids() == ml_ds.items.ids())


def test_unrated_selector(ml_ds: Dataset):
    sel = UnratedTrainingItemsCandidateSelector()
    sel.train(ml_ds)

    row = ml_ds.user_row(100)
    assert row is not None
    cands = sel(query=row)

    assert len(cands) <= ml_ds.item_count
    assert len(cands) == len(set(ml_ds.items.ids()) - set(row.ids()))


def test_unrated_selector_many(rng: np.random.Generator, ml_ds: Dataset):
    sel = UnratedTrainingItemsCandidateSelector()
    sel.train(ml_ds)

    for u in rng.choice(ml_ds.users.ids(), 200):
        row = ml_ds.user_row(100)
        assert row is not None
        cands = sel(query=row)

        assert len(cands) <= ml_ds.item_count
        assert len(cands) == len(set(ml_ds.items.ids()) - set(row.ids()))
        assert not np.any(np.isin(cands.ids(), row.ids()))
        assert not np.any(np.isin(cands.numbers(), row.numbers()))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Literal

import click

from lenskit.data import ItemListCollection
from lenskit.data.dataset import Dataset
from lenskit.logging import get_logger
from lenskit.random import random_generator

_log = get_logger(__name__)


@click.command("subset")
@click.option(
    "--item-lists", is_flag=True, help="Write an ItemListCollection instead of a Dataset."
)
@click.option(
    "--layout",
    type=click.Choice(["native", "flat"]),
    default="native",
    help="Specify layout for item list output.",
)
@click.option(
    "--sample-rows",
    type=int,
    metavar="N",
    help="Subset to at most N rows of the interaction data",
)
@click.argument("src", required=True, type=Path)
@click.argument("dst", required=True, type=Path)
def subset(
    src: Path,
    dst: Path,
    sample_rows: int | None,
    item_lists: bool,
    layout: Literal["native", "flat"],
):
    """
    Subset a LensKit dataset.

    The input SRC can be either a LensKit native dataset or a Parquet file
    containing an item list colleciton in native format.  The "rows" in this
    subsetting operation are entire rows of the interaction matrix, typically
    users.
    """
    log = _log.bind(src=str(src))

    dataset: Dataset | None = None
    lists: ItemListCollection
    if src.is_file():
        log.info("loading ItemListCollection")
        lists = ItemListCollection.load_parquet(src)
    else:
        log.debug("loading Dataset")
        dataset = Dataset.load(src)
        log.info("extracting default interactions")
        if len(dataset.schema.relationships) > 1:
            log.warn("Dataset has multiple relationships, only default interactions are subset")

        # TODO: support non-default interactions / multiple interaction sets
        lists = dataset.interactions().matrix().to_ilc()

    if sample_rows is not None:
        lists = _subset_lists(lists, sample_rows)

    log = _log.bind(dst=str(dst))
    if item_lists:
        log.info("saving ItemListCollection Parquet in %s layout", layout)
        lists.save_parquet(dst, layout=layout)
    else:
        log.info("saving full dataset")
        lists.to_dataset().save(dst)


def _subset_lists(lists: ItemListCollection, n: int) -> ItemListCollection:
    """
    Subset the user lists.
    """
    log = _log.bind(n_input=len(lists), n_output=n)
    if len(lists) <= n:
        log.warn("input count exceeds target count, keeping all")
        return lists

    rng = random_generator()

    log.info("subsetting lists")
    log.debug("picking row indices")
    rows = rng.choice(len(lists), n, replace=True)

    log.debug("converting to arrow table")
    table = lists.to_arrow()
    log.debug("subsetting arrow table")
    table = table.take(rows)
    log.debug("converting back to item list collection")
    return ItemListCollection.from_arrow(table)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from pytest import raises

from lenskit.pipeline import PipelineState


def test_empty():
    state = PipelineState()
    assert len(state) == 0
    assert not state
    assert "scroll" not in state

    with raises(KeyError):
        state["scroll"]


def test_single_value():
    state = PipelineState({"scroll": "HACKEM MUCHE"})
    assert len(state) == 1
    assert state
    assert "scroll" in state
    assert state["scroll"] == "HACKEM MUCHE"


def test_alias():
    state = PipelineState({"scroll": "HACKEM MUCHE"}, {"book": "scroll"})
    assert len(state) == 1
    assert state
    assert "scroll" in state
    assert "book" in state
    assert state["book"] == "HACKEM MUCHE"


def test_alias_missing():
    state = PipelineState({"scroll": "HACKEM MUCHE"}, {"book": "manuscript"})
    assert len(state) == 1
    assert state
    assert "scroll" in state
    assert "book" not in state



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd
from numpy.typing import NDArray

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given, settings

from lenskit.basic import PopScorer
from lenskit.basic.history import KnownRatingScorer, UserTrainingHistoryLookup
from lenskit.basic.topn import TopNRanker
from lenskit.data import Dataset, ItemList, RecQuery

_log = logging.getLogger(__name__)


def test_lookup_user_id(ml_ds):
    lookup = UserTrainingHistoryLookup()
    lookup.train(ml_ds)

    user = ml_ds.users.id(100)
    ds_row = ml_ds.user_row(user)
    query = lookup(user)

    assert isinstance(query, RecQuery)
    assert query.user_id == user
    assert query.user_items is not None
    assert len(query.user_items) == len(ds_row)
    assert np.all(query.user_items.ids() == ds_row.ids())


def test_lookup_no_override(ml_ds):
    "ensure history does not override when items already present"
    lookup = UserTrainingHistoryLookup()
    lookup.train(ml_ds)

    user = ml_ds.users.id(100)
    ds_row = ml_ds.user_row(user)
    query = RecQuery(user, ds_row[:-2])
    query = lookup(query)

    assert isinstance(query, RecQuery)
    assert query.user_id == user
    assert query.user_items is not None
    assert len(query.user_items) == len(ds_row) - 2
    assert np.all(query.user_items.ids() == ds_row[:-2].ids())


def test_lookup_items_only(ml_ds: Dataset):
    "ensure history accepts item list"
    lookup = UserTrainingHistoryLookup()
    lookup.train(ml_ds)

    user = ml_ds.users.id(100)
    ds_row = ml_ds.user_row(user)
    query = lookup(ds_row[:-5])

    assert isinstance(query, RecQuery)
    assert query.user_id is None
    assert query.user_items is not None
    assert len(query.user_items) == len(ds_row) - 5
    assert np.all(query.user_items.ids() == ds_row[:-5].ids())


def test_lookup_pickle(ml_ds: Dataset):
    "ensure we can correctly pickle a history component"
    lookup = UserTrainingHistoryLookup()
    lookup.train(ml_ds)

    blob = pickle.dumps(lookup)
    l2 = pickle.loads(blob)
    assert isinstance(l2, UserTrainingHistoryLookup)

    assert l2.interactions.count() == lookup.interactions.count()

    ds_row = ml_ds.user_row(user_id=100)
    l_row = lookup(100)
    l2_row = l2(100)

    assert l_row.user_id == 100
    assert np.all(l_row.user_items.ids() == ds_row.ids())
    assert l2_row.user_id == 100
    assert np.all(l2_row.user_items.ids() == ds_row.ids())


def test_known_rating_defaults(ml_ds: Dataset):
    algo = KnownRatingScorer()
    algo.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = algo(2, ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)

    scores = scored.scores()
    assert scores is not None
    assert np.all(np.isnan(scores[-3:]))
    assert not np.any(np.isnan(scores[:-3]))

    row = ml_ds.user_row(2)
    assert row is not None
    udf = row.field("rating", "pandas", index="ids")
    assert udf is not None
    assert np.all(scores[~np.isnan(scores)] == udf.loc[[50, 17]].values)


def test_known_indicator(ml_ds: Dataset):
    algo = KnownRatingScorer(score="indicator")
    algo.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = algo(2, ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)

    scores = scored.scores()
    assert scores is not None
    assert np.all(scores == [1, 1, 0, 0, 0])


def test_known_query_ratings_none(ml_ds: Dataset):
    algo = KnownRatingScorer(score="rating", source="query")
    algo.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = algo(2, ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)

    # but no scores should be computed!
    scores = scored.scores()
    assert scores is not None
    assert np.all(np.isnan(scores))


def test_known_query_ratings(ml_ds: Dataset):
    algo = KnownRatingScorer(score="rating", source="query")
    algo.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = algo(ml_ds.user_row(2), ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)

    scores = scored.scores()
    assert scores is not None
    assert np.all(np.isnan(scores[-3:]))
    assert not np.any(np.isnan(scores[:-3]))

    row = ml_ds.user_row(2)
    assert row is not None
    udf = row.field("rating", "pandas", index="ids")
    assert udf is not None
    assert np.all(scores[~np.isnan(scores)] == udf.loc[[50, 17]].values)


def test_known_query_indicator(ml_ds: Dataset):
    algo = KnownRatingScorer(score="indicator", source="query")
    algo.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = algo(ml_ds.user_row(2), ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)

    scores = scored.scores()
    assert scores is not None
    assert np.all(scores == [1, 1, 0, 0, 0])


def test_known_query_default_indicator(ml_ds: Dataset):
    algo = KnownRatingScorer(source="query")
    algo.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = algo(ItemList(ml_ds.user_row(2), rating=False), ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)

    scores = scored.scores()
    assert scores is not None
    assert np.all(scores[:2] == [1, 1])
    assert np.all(np.isnan(scores[2:]))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit logging processors and converters.
"""

import multiprocessing as mp
from collections import deque
from datetime import datetime
from typing import Any

import structlog
from structlog.typing import EventDict

LOGGED_ERRORS = deque([], 5)


def filter_exceptions(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    exc = event_dict.get("exc_info", None)
    if isinstance(exc, Exception):
        count = getattr(exc, "_lenskit_seen", 0)
        exc._lenskit_seen = count + 1
        if count > 1:
            del event_dict["exc_info"]

    return event_dict


def remove_internal(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    """
    Filter out “internal” attrs (beginning with ``_``) for console logging.
    """

    to_del = [k for k in event_dict.keys() if k.startswith("_")]
    for k in to_del:
        del event_dict[k]

    return event_dict


def format_timestamp(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    """
    Reformat UNIX timestamps.
    """

    if "timestamp" in event_dict:
        stamp = datetime.fromtimestamp(event_dict["timestamp"])
        event_dict = dict(event_dict)
        event_dict["timestamp"] = stamp.isoformat(timespec="seconds")
        return event_dict
    else:
        return event_dict


def add_process_info(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    """
    Add process info if it does not exist.
    """

    proc = mp.current_process()
    if "pid" not in event_dict:
        event_dict["pid"] = proc.pid
    if "pname" not in event_dict:
        event_dict["pname"] = proc.name

    return event_dict


def log_warning(message, category, filename, lineno, file=None, line=None):
    log = structlog.stdlib.get_logger("lenskit")
    log = log.bind(category=category.__name__, file=filename, lineno=lineno)
    log.warning(str(message))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

import pytest

from lenskit.data import ItemList
from lenskit.metrics.ranking import AveragePrecision


def _test_ap(items, rel, **k):
    recs = ItemList(items, ordered=True)
    truth = ItemList(rel)
    return AveragePrecision(**k).measure_list(recs, truth)


def test_ap_empty_none():
    ap = _test_ap([], [1, 2, 3])
    assert np.isnan(ap)

    ap = _test_ap([], [10, 20, 30])
    assert np.isnan(ap)


def test_ap_all_relevant():
    ap = _test_ap([1, 3, 5], [1, 3, 5])
    assert ap == pytest.approx(1.0)

    ap = _test_ap([1, 5, 7], [1, 5, 7])
    assert ap == pytest.approx(1.0)


def test_ap_partial_hits():
    ap = _test_ap([1, 2, 3], [1, 3])
    assert ap == pytest.approx((1.0 + 2 / 3) / 2)

    ap = _test_ap([1, 2, 3], [2, 4])
    assert ap == pytest.approx(0.25)


def test_ap_cutoff_k():
    ap = _test_ap([1, 2, 3, 4, 5], [1, 3, 5], k=3)
    assert ap == pytest.approx((1.0 + 2 / 3) / 3)

    ap = _test_ap([1, 2, 3, 4, 5], [4, 5, 6], k=2)
    assert ap == pytest.approx(0.0)


def test_ap_no_hits():
    ap = _test_ap([10, 20, 30], [1, 2, 3])
    assert ap == pytest.approx(0.0)

    ap = _test_ap([1, 4, 6], [2, 3, 5])
    assert ap == pytest.approx(0.0)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Singular value decomposition for explicit feedback.

This module contains a truncated SVD explicit-feedback scorer built on
:class:`sklearn.decomposition.TruncatedSVD`.
"""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from pydantic import AliasChoices, BaseModel, Field
from sklearn.decomposition import TruncatedSVD
from typing_extensions import Literal, override

from lenskit.basic import BiasModel, Damping
from lenskit.config.common import EmbeddingSizeMixin
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery
from lenskit.data.vocab import Vocabulary
from lenskit.logging import Stopwatch, get_logger
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


class BiasedSVDConfig(EmbeddingSizeMixin, BaseModel):
    embedding_size: int = Field(
        default=50, validation_alias=AliasChoices("embedding_size", "features")
    )
    """
    The dimension of user and item embeddings (number of latent features to
    learn).
    """

    damping: Damping = 5
    algorithm: Literal["arpack", "randomized"] = "randomized"
    n_iter: int = 5


class BiasedSVDScorer(Component[ItemList], Trainable):
    """
    Biased matrix factorization for explicit feedback using SciKit-Learn's
    :class:`~sklearn.decomposition.TruncatedSVD`.  It operates by first
    computing the bias, then computing the SVD of the bias residuals.

    You'll generally want one of the iterative SVD implementations such as
    :class:`lenskit.als.BiasedMFScorer`; this is here primarily as
    an example and for cases where you want to evaluate a pure SVD
    implementation.

    Stability:
        Caller
    """

    config: BiasedSVDConfig

    bias_: BiasModel
    factorization_: TruncatedSVD
    users_: Vocabulary
    items_: Vocabulary
    user_components_: NDArray[np.float64]

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "factorization_") and not options.retrain:
            return

        timer = Stopwatch()
        _log.info("[%s] computing bias", timer)
        self.bias_ = BiasModel.learn(data, self.config.damping)

        g_bias = self.bias_.global_bias
        u_bias = self.bias_.user_biases
        i_bias = self.bias_.item_biases

        _log.info("[%s] sparsifying and normalizing matrix", timer)
        r_mat = data.interaction_matrix(format="scipy", field="rating", layout="coo", legacy=True)
        # copy the data and start subtracting
        r_mat.data = r_mat.data - g_bias
        if i_bias is not None:
            r_mat.data -= i_bias[r_mat.col]
        if u_bias is not None:
            r_mat.data -= u_bias[r_mat.row]

        r_mat = r_mat.tocsr()

        self.factorization_ = TruncatedSVD(
            self.config.embedding_size, algorithm=self.config.algorithm, n_iter=self.config.n_iter
        )
        _log.info("[%s] training SVD (k=%d)", timer, self.factorization_.n_components)  # type: ignore
        Xt = self.factorization_.fit_transform(r_mat)  # type: ignore
        self.user_components_ = Xt
        self.users_ = data.users
        self.items_ = data.items
        _log.info("finished model training in %s", timer)

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        uidx = None
        if query.user_id is not None:
            uidx = self.users_.number(query.user_id, missing="none")

        if uidx is None:
            return ItemList(items, scores=np.nan)

        # Get index for user & usable items
        iidx = items.numbers(vocabulary=self.items_, missing="negative")
        good_iidx = iidx[iidx >= 0]

        _log.debug("reverse-transforming user %s (idx=%d)", query.user_id, uidx)
        Xt = self.user_components_[[uidx], :]
        X = self.factorization_.inverse_transform(Xt)
        # restrict to usable desired items
        Xsel = X[0, good_iidx]

        scores = np.full(len(items), np.nan)
        scores[iidx >= 0] = Xsel

        biases, _ub = self.bias_.compute_for_items(items, query.user_id, query.user_items)
        scores += biases

        return ItemList(items, scores=scores)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
from itertools import product

import numpy as np
import pandas as pd
import torch

from pytest import approx, mark

from lenskit.data import Dataset, ItemList, RecQuery, from_interactions_df, load_movielens_df
from lenskit.flexmf import FlexMFImplicitConfig, FlexMFImplicitScorer
from lenskit.graphs.lightgcn import LightGCNConfig, LightGCNScorer
from lenskit.metrics import quick_measure_model
from lenskit.testing import BasicComponentTests, ScorerTests


class TestLightGCNPairwise(BasicComponentTests, ScorerTests):
    component = LightGCNScorer
    config = LightGCNConfig(loss="pairwise")


class TestLightGCNLogistic(BasicComponentTests, ScorerTests):
    component = LightGCNScorer
    config = LightGCNConfig(loss="logistic")


@mark.slow
@mark.eval
def test_lightgcn_test_accuracy(ml_100k):
    ds = from_interactions_df(ml_100k)
    results = quick_measure_model(
        LightGCNScorer(embedding_size=25, epochs=10, batch_size=1024),
        ds,
    )

    print(results.list_summary())

    assert results.list_summary().loc["RBP", "mean"] >= 0.01
    assert results.list_summary().loc["RBP", "mean"] < 0.25



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Formatting utilities for log and diagnostic output.
"""

from datetime import timedelta


def friendly_duration(elapsed: float | timedelta):
    """
    Short, human-friendly representation of a duration.
    """
    if isinstance(elapsed, timedelta):
        elapsed = elapsed.total_seconds()

    if elapsed < 1:
        return "{: 0.0f}ms".format(elapsed * 1000)
    elif elapsed > 60 * 60:
        h, m = divmod(elapsed, 60 * 60)
        m, s = divmod(m, 60)
        return "{:0.0f}h{:0.0f}m{:0.2f}s".format(h, m, s)
    elif elapsed > 60:
        m, s = divmod(elapsed, 60)
        return "{:0.0f}m{:0.2f}s".format(m, s)
    else:
        return "{:0.2f}s".format(elapsed)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given
from pytest import raises

from lenskit.data.checks import check_1d, check_type


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes(min_dims=1, max_dims=1)))
def test_check_1d_ok(arr):
    check_1d(arr)


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes(min_dims=1, max_dims=1)))
def test_check_1d_ok_return(arr):
    assert check_1d(arr, error="return")


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes(min_dims=2)))
def test_check_1d_bad(arr):
    with raises(TypeError, match="must be 1D"):
        check_1d(arr)


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes(min_dims=2)))
def test_check_1d_bad_return(arr):
    assert not check_1d(arr, error="return")


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes()), st.integers(min_value=0))
def test_check_expected_size(arr, exp):
    if arr.shape == (exp,):
        check_1d(arr, exp)
    else:
        with raises(TypeError):
            check_1d(arr, exp)


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes()), st.integers(min_value=0))
def test_check_expected_size_return(arr, exp):
    if arr.shape == (exp,):
        assert check_1d(arr, exp, error="return")
    else:
        assert not check_1d(arr, exp, error="return")


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes()))
def test_check_type_ok(arr):
    check_type(arr, arr.dtype.type)


@given(nph.arrays(nph.floating_dtypes(), nph.array_shapes()))
def test_check_type_ok_subclass(arr):
    check_type(arr, np.floating)


@given(nph.arrays(st.one_of(nph.integer_dtypes(), nph.floating_dtypes()), nph.array_shapes()))
def test_check_type_ok_multi(arr):
    check_type(arr, np.integer, np.floating)


@given(nph.arrays(nph.scalar_dtypes(), nph.array_shapes()))
def test_check_type_ok_return(arr):
    assert check_type(arr, arr.dtype.type, error="return")


@given(nph.arrays(nph.floating_dtypes(), nph.array_shapes()))
def test_check_type_bad_float(arr):
    with raises(TypeError):
        check_type(arr, np.integer)


@given(nph.arrays(nph.floating_dtypes(), nph.array_shapes()))
def test_check_type_bad_float_return(arr):
    assert not check_type(arr, np.integer, error="return")


@given(nph.arrays(nph.integer_dtypes(), nph.array_shapes()))
def test_check_type_bad_int(arr):
    with raises(TypeError):
        check_type(arr, np.floating)


@given(nph.arrays(nph.unicode_string_dtypes(), nph.array_shapes()))
def test_check_type_bad_str(arr):
    with raises(TypeError):
        check_type(arr, np.number)

    with raises(TypeError):
        check_type(arr, np.integer, np.floating)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
from os import fspath
from pathlib import Path

from click.testing import CliRunner

from lenskit.cli import lenskit
from lenskit.testing import ml_test_dir


def test_data_describe_ml():
    runner = CliRunner()
    result = runner.invoke(lenskit, ["data", "describe", "--movielens", fspath(ml_test_dir)])

    assert result.exit_code == 0

    summary = result.output
    print(summary)
    assert "Summary of ml-latest-small" in summary
    assert "tag_counts" in summary


def test_data_describe_ml_markdown():
    runner = CliRunner()
    result = runner.invoke(
        lenskit, ["data", "describe", "--movielens", "--markdown", fspath(ml_test_dir)]
    )

    assert result.exit_code == 0

    summary = result.output
    print(summary)
    assert "# Summary of ml-latest-small" in summary
    assert "tag_counts" in summary



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Nonnegative matrix factorization for implicit feedback.

This module contains a non-negative factorization implicit-feedback scorer built on
:func:`sklearn.decomposition.non_negative_factorization`.
"""

from __future__ import annotations

import numpy as np
from pydantic import AliasChoices, BaseModel, Field, PositiveInt
from sklearn.decomposition import non_negative_factorization
from typing_extensions import Literal, override

from lenskit.config.common import EmbeddingSizeMixin
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery
from lenskit.data.types import NPMatrix
from lenskit.data.vocab import Vocabulary
from lenskit.logging import Stopwatch, get_logger
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


class NMFConfig(EmbeddingSizeMixin, BaseModel, extra="forbid"):
    """
    Configuration for :class:`NMFScorer`.
    See the documentation for :func:`sklearn.decomposition.non_negative_factorization`
    for the configuration options.

    """

    beta_loss: Literal["frobenius", "kullback-leibler", "itakura-saito"] = "frobenius"
    max_iter: PositiveInt = Field(default=200, validation_alias=AliasChoices("max_iter", "epochs"))
    embedding_size: PositiveInt | None = Field(
        default=None, validation_alias=AliasChoices("embedding_size", "n_components")
    )
    alpha_W: float = 0.0
    alpha_H: float | Literal["same"] = "same"
    l1_ratio: float = 0.0


class NMFScorer(Component[ItemList], Trainable):
    """
    Non-negative matrix factorization for implicit feedback using SciKit-Learn's
    :func:`sklearn.decomposition.non_negative_factorization`. It computes the user
    and item embedding matrices using an indicator matrix as the input.

    Stability:
        Caller
    """

    config: NMFConfig

    users: Vocabulary
    items: Vocabulary
    user_components: NPMatrix
    item_components: NPMatrix

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_components") and not options.retrain:
            return

        timer = Stopwatch()

        _log.info("[%s] sparsifying and normalizing matrix", timer)
        r_mat = data.interactions().matrix().scipy(layout="csr", legacy=True)

        _log.info("[%s] training NMF", timer)
        W, H, n_iter = non_negative_factorization(
            r_mat,
            beta_loss=self.config.beta_loss,
            max_iter=self.config.max_iter,
            n_components=self.config.embedding_size,
            alpha_W=self.config.alpha_W,
            alpha_H=self.config.alpha_H,
            l1_ratio=self.config.l1_ratio,
        )
        _log.info("[%s] Trained NMF in %d iterations", timer, n_iter)

        self.user_components = np.require(W, dtype=np.float32)
        self.item_components = np.require(H.T, dtype=np.float32)
        self.users = data.users
        self.items = data.items
        _log.info("finished model training in %s", timer)

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        uidx = None
        if query.user_id is not None:
            uidx = self.users.number(query.user_id, missing="none")

        if uidx is None:
            return ItemList(items, scores=np.nan)

        # Get index for user & usable items
        iidx = items.numbers(vocabulary=self.items, missing="negative")
        good_iidx = iidx[iidx >= 0]

        _log.debug("reverse-transforming user %s (idx=%d)", query.user_id, uidx)
        W = self.user_components[[uidx], :]
        H = self.item_components[good_iidx, :]
        S = W @ H.T

        scores = np.full(len(items), np.nan)
        scores[iidx >= 0] = S[0, :]

        return ItemList(items, scores=scores)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
User-based k-NN collaborative filtering.
"""

# pyright: basic
from __future__ import annotations

import warnings

import numpy as np
import pandas as pd
import pyarrow as pa
import scipy.sparse.linalg as spla
from pydantic import AliasChoices, BaseModel, Field, PositiveFloat, PositiveInt, field_validator
from scipy.sparse import csr_array
from typing_extensions import NamedTuple, Optional, override

from lenskit._accel import knn
from lenskit.data import Dataset, FeedbackType, ItemList, QueryInput, RecQuery
from lenskit.data.matrix import SparseRowArray
from lenskit.data.vocab import Vocabulary
from lenskit.diagnostics import DataWarning
from lenskit.logging import Stopwatch, get_logger
from lenskit.parallel.config import ensure_parallel_init
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


class UserKNNConfig(BaseModel, extra="forbid"):
    "Configuration for :class:`ItemKNNScorer`."

    max_nbrs: PositiveInt = Field(20, validation_alias=AliasChoices("max_nbrs", "nnbrs", "k"))
    """
    The maximum number of neighbors for scoring each item.
    """
    min_nbrs: PositiveInt = 1
    """
    The minimum number of neighbors for scoring each item.
    """
    min_sim: PositiveFloat = 1.0e-6
    """
    Minimum similarity threshold for considering a neighbor.  Must be positive;
    if less than the smallest 32-bit normal (:math:`1.175 \\times 10^{-38}`), is
    clamped to that value.
    """
    feedback: FeedbackType = "explicit"
    """
    The type of input data to use (explicit or implicit).  This affects data
    pre-processing and aggregation.
    """

    @field_validator("min_sim", mode="after")
    @staticmethod
    def clamp_min_sim(sim) -> float:
        return max(sim, float(np.finfo(np.float64).smallest_normal))

    @property
    def explicit(self) -> bool:
        """
        Query whether this is in explicit-feedback mode.
        """
        return self.feedback == "explicit"


class UserKNNScorer(Component[ItemList], Trainable):
    """
    User-user nearest-neighbor collaborative filtering with ratings. This
    user-user implementation is not terribly configurable; it hard-codes design
    decisions found to work well in the previous Java-based LensKit code.

    .. note::

        This component must be used with queries containing the user's history,
        either directly in the input or by wiring its query input to the output of a
        user history component (e.g., :class:`~lenskit.basic.UserTrainingHistoryLookup`).

    Stability:
        Caller
    """

    config: UserKNNConfig

    users: Vocabulary
    "The index of user IDs."
    items: Vocabulary
    "The index of item IDs."
    user_means: np.ndarray[tuple[int], np.dtype[np.float32]] | None
    "Mean rating for each known user."
    user_vectors: csr_array
    "Normalized rating matrix (CSR) to find neighbors at prediction time."
    user_ratings: SparseRowArray
    "Centered but un-normalized rating matrix (COO) to find neighbor ratings."

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        "Train" a user-user CF model.  This memorizes the rating data in a format that is usable
        for future computations.

        Args:
            ratings(pandas.DataFrame): (user, item, rating) data for collaborative filtering.
        """
        if hasattr(self, "user_ratings_") and not options.retrain:
            return

        ensure_parallel_init()
        rmat = (
            data.interactions().matrix().scipy(attribute="rating" if self.config.explicit else None)
        ).astype(np.float32)

        rmat, means = self._center_ratings(rmat)
        normed = self._normalize_rows(rmat)

        self.user_vectors = normed
        self.user_ratings = SparseRowArray.from_scipy(rmat, values=self.config.explicit)
        self.users = data.users
        self.user_means = means
        self.items = data.items

    def _center_ratings(self, rmat: csr_array) -> tuple[csr_array, np.ndarray | None]:
        if self.config.explicit:
            counts = np.diff(rmat.indptr)
            sums = rmat.sum(axis=1)
            means = np.zeros(sums.shape, dtype=np.float32)
            # manual divide to avoid division by zero
            np.divide(sums, counts, out=means, where=counts > 0)
            rmat.data = rmat.data - np.repeat(means, counts)
            if np.allclose(rmat.data, 0.0):
                warnings.warn(
                    "Ratings seem to have the same value, centering is not recommended.",
                    DataWarning,
                )
            return rmat, means
        else:
            return rmat, None

    def _normalize_rows(self, rmat: csr_array) -> csr_array:
        norms = spla.norm(rmat, 2, axis=1)
        assert norms.shape == (rmat.shape[0],)
        # clamp small values to avoid divide by 0 (only appear when an entry is all 0)
        cmat = rmat / np.maximum(norms, np.finfo("f4").smallest_normal).reshape(-1, 1)
        assert cmat.shape == rmat.shape
        return cmat.tocsr()

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Compute predictions for a user and items.

        Args:
            user: the user ID
            items (array-like): the items to predict
            ratings (pandas.Series):
                the user's ratings (indexed by item id); if provided, will be used to
                recompute the user's bias at prediction time.

        Returns:
            pandas.Series: scores for the items, indexed by item id.
        """
        query = RecQuery.create(query)
        watch = Stopwatch()
        log = _log.bind(user_id=query.user_id, n_items=len(items))
        if len(items) == 0:
            log.debug("no candidate items, skipping")
            return ItemList(items, scores=np.nan)

        udata = self._get_user_data(query)
        if udata is None:
            log.debug("user has no ratings, skipping")
            return ItemList(items, scores=np.nan)

        uidx, ratings, umean = udata
        assert ratings.shape == (len(self.items),)  # ratings is a dense vector

        # now ratings has vbeen normalized to be a mean-centered unit vector
        # this means we can dot product to score neighbors
        # score the neighbors!
        nbr_sims = self.user_vectors @ ratings
        assert nbr_sims.shape == (len(self.users),)
        if uidx is not None:
            # zero out the self-similarity
            nbr_sims[uidx] = 0

        # get indices for these neighbors
        nbr_idxs = np.arange(len(self.users), dtype=np.int32)

        nbr_mask = nbr_sims >= self.config.min_sim

        kn_sims = nbr_sims[nbr_mask]
        kn_idxs = nbr_idxs[nbr_mask]
        if len(kn_sims) > 0:
            log.debug(
                "found %d candidate neighbors (of %d total), max sim %0.4f",
                len(kn_sims),
                len(self.users),
                np.max(kn_sims).item(),
            )
        else:
            log.debug("no candidate neighbors found, cannot score")
            return ItemList(items, scores=np.nan)

        assert not np.any(np.isnan(kn_sims))

        iidxs = items.numbers(vocabulary=self.items, missing="negative")

        ki_mask = iidxs >= 0
        usable_iidxs = iidxs[ki_mask]

        usable_iidxs = pa.array(usable_iidxs, pa.int32())
        kn_idxs = pa.array(kn_idxs, pa.int32())
        kn_sims = pa.array(kn_sims, pa.float32())

        if self.config.explicit:
            scores = knn.user_score_items_explicit(
                usable_iidxs,
                kn_idxs,
                kn_sims,
                self.user_ratings,
                self.config.max_nbrs,
                self.config.min_nbrs,
            )
        else:
            scores = knn.user_score_items_implicit(
                usable_iidxs,
                kn_idxs,
                kn_sims,
                self.user_ratings,
                self.config.max_nbrs,
                self.config.min_nbrs,
            )

        scores = scores.to_numpy(zero_copy_only=False, writable=True)
        scores += umean

        results = pd.Series(scores, index=items.ids()[ki_mask], name="prediction")
        results = results.reindex(items.ids())

        log.debug(
            "scored %d items in %s",
            results.notna().sum(),
            watch,
        )
        return ItemList(items, scores=results.values)  # type: ignore

    def _get_user_data(self, query: RecQuery) -> Optional[UserRatings]:
        "Get a user's data for user-user CF"

        index = self.users.number(query.user_id, missing=None)

        if query.user_items is None:
            if index is None:
                _log.warning("user %s has no ratings and none provided", query.user_id)
                return None

            index = int(index)
            row = self.user_vectors[[index], :].toarray()[0, :]
            if self.config.explicit:
                assert self.user_means is not None
                umean = self.user_means[index].item()
            else:
                umean = 0
            return UserRatings(index, row, umean)
        elif len(query.user_items) == 0:
            return None
        else:
            _log.debug("using provided item history")
            ratings = np.zeros(len(self.items), dtype=np.float32)
            ui_nos = query.user_items.numbers(missing="negative", vocabulary=self.items)
            ui_mask = ui_nos >= 0

            if self.config.explicit:
                urv = query.user_items.field("rating")
                if urv is None:
                    _log.warning("user %s has items but no ratings", query.user_id)
                    return None

                urv = np.require(urv, dtype=np.float32)
                umean = urv.mean().item()
                ratings[ui_nos[ui_mask]] = urv[ui_mask] - umean
            else:
                umean = 0
                ratings[ui_nos[ui_mask]] = 1.0

            return UserRatings(index, ratings, umean)


class UserRatings(NamedTuple):
    """
    Dense user ratings.
    """

    index: int | None
    ratings: np.ndarray[tuple[int], np.dtype[np.float32]]
    mean: float



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import approx, fixture, mark, raises

from lenskit.basic import PopScorer
from lenskit.data import ItemList, ItemListCollection
from lenskit.metrics import NDCG, Recall
from lenskit.metrics._accum import MetricAccumulator, MetricWrapper
from lenskit.metrics._base import DecomposedMetric, GlobalMetric, ListMetric, Metric
from lenskit.metrics.basic import ListLength
from lenskit.splitting import split_temporal_fraction

_log = logging.getLogger(__name__)


@fixture
def basic_accumulator():
    acc = MetricAccumulator()
    acc.add_metric(Recall(5))
    acc.add_metric(NDCG(5))
    return acc


@fixture
def sample_lists():
    return {
        "recs1": ItemList([1, 2, 3], user=["u1"] * 3),
        "test1": ItemList([2, 3]),
        "recs2": ItemList([4, 5], user=["u2"] * 2),
        "test2": ItemList([5]),
    }


def test_accumulator_empty_and_unmeasured_defaults():
    acc = MetricAccumulator()
    # initial state
    assert acc.metrics == []
    assert acc.list_metrics().empty
    assert not acc.summary_metrics()

    # measuring with no metrics
    acc.measure_list(ItemList([1]), ItemList([1]), user="u1")
    assert acc.list_metrics().empty
    assert not acc.summary_metrics()

    # add metrics but do not measure
    acc.add_metric(Recall(5))
    acc.add_metric(NDCG(5))
    assert not acc.summary_metrics()


# measuring lists


def test_accumulator_measures_list_and_summary(sample_lists):
    acc = MetricAccumulator()
    acc.add_metric(ListLength())

    acc.measure_list(sample_lists["recs1"], sample_lists["test1"], user="u1")
    acc.measure_list(sample_lists["recs2"], sample_lists["test2"], user="u2")

    list_metrics = acc.list_metrics()
    summary = acc.summary_metrics()

    assert len(list_metrics) == 2
    assert set(list_metrics["N"]) == {2.0, 3.0}
    assert approx(summary["N"]) == 2.5


def test_accumulator_empty_itemlists():
    acc = MetricAccumulator()
    acc.add_metric(ListLength())

    acc.measure_list(ItemList([]), ItemList([1, 2]), user="u1")
    acc.measure_list(ItemList([1, 2]), ItemList([]), user="u2")
    acc.measure_list(ItemList([]), ItemList([]), user="u3")

    metrics = acc.list_metrics()
    assert len(metrics) == 3
    assert all(metrics["N"] >= 0)


def test_list_metrics_no_key_fields():
    acc = MetricAccumulator()
    acc.add_metric(ListLength())
    acc.measure_list(ItemList([1, 2]), ItemList([1]))
    metrics = acc.list_metrics()
    assert len(metrics) == 1
    assert metrics.index.names == [None]


# metric wrapping


def test_wrap_metric_with_class():
    acc = MetricAccumulator()
    wrapper = acc._wrap_metric(ListLength, label="ListLength", default=None)
    assert isinstance(wrapper.metric, ListLength)
    assert wrapper.label == "ListLength"


def test_wrap_metric_function_label():
    def custom_metric(recs, test):
        return len(recs)

    acc = MetricAccumulator()
    wrapper = acc._wrap_metric(custom_metric, None, None)
    assert wrapper.label == "function"


# key fields and duplicates


@mark.parametrize(
    "keys,expected_names",
    [
        ({"user": "u1"}, ["user"]),
        ({"user": "u1", "fold": 1}, ["user", "fold"]),
        ({"user": "u1", "session": "a", "fold": 1}, ["user", "session", "fold"]),
    ],
)
def test_accumulator_key_fields(keys, expected_names):
    acc = MetricAccumulator()
    acc.add_metric(ListLength())
    acc.measure_list(ItemList([1, 2]), ItemList([2]), **keys)
    assert acc._key_fields == expected_names
    metrics = acc.list_metrics()
    assert set(metrics.index.names) == set(expected_names)


# custom metrics and types


@mark.parametrize(
    "metric_input,expected_type",
    [
        (ListLength(), ListLength),
        (ListLength, ListLength),
        (lambda r, t: len(r), type(lambda: None)),
    ],
)
def test_add_metric_types(metric_input, expected_type):
    acc = MetricAccumulator()
    acc.add_metric(metric_input)
    assert isinstance(acc.metrics[0].metric, expected_type)


def test_custom_labels_and_defaults():
    acc = MetricAccumulator()
    acc.add_metric(ListLength(), label="CustomLength")
    assert acc.metrics[0].label == "CustomLength"
    acc.add_metric(ListLength(), default=99.0)
    assert acc.metrics[1].default == 99.0


# global, callable, scalar metrics


def test_global_metric():
    class DummyGlobalMetric(GlobalMetric):
        label = "dummy_global"

        def measure_run(self, run, test):
            return 42.0

    acc = MetricAccumulator()
    acc.add_metric(DummyGlobalMetric())
    result = acc.metrics[0].measure_run(ItemListCollection([]), ItemListCollection([]))
    assert result == 42.0
    assert acc.metrics[0].is_global
    assert not acc.metrics[0].is_listwise


def test_callable_metric():
    def _callable(recs, test):
        return len(recs)

    acc = MetricAccumulator()
    acc.add_metric(_callable, label="callable_metric")
    recs = ItemList([1, 2, 3])
    test_il = ItemList([1])
    result = acc.metrics[0].measure_list(recs, test_il)
    assert result == 3
    wrapper = acc.metrics[0]
    assert wrapper.is_listwise
    assert not wrapper.is_global


def test_scalar_metric():
    class ScalarMetric(Metric):
        label = "scalar"

        def measure_list(self, recs, test):
            return 5

        def summarize(self, values):
            return 99.0

    acc = MetricAccumulator()
    acc.add_metric(ScalarMetric())
    acc.measure_list(ItemList([1]), ItemList([1]), user="u1")
    summary = acc.summary_metrics()
    assert summary["scalar"] == 99.0


# metrics returning mixed


def test_mixed_metric_behavior():
    class MixedMetric(Metric):
        label = "Mixed"

        def __init__(self):
            self.call_count = 0

        def measure_list(self, recs, test):
            self.call_count += 1
            return 1.0 if self.call_count % 2 == 1 else {"b": 2.0}

        def extract_list_metrics(self, data):
            return data

        def summarize(self, values):
            numeric_vals = [v if isinstance(v, float) else v["b"] for v in values]
            return {"mean": np.mean(numeric_vals)}

    acc = MetricAccumulator()
    acc.add_metric(MixedMetric())
    acc.measure_list(ItemList([1]), ItemList([1]), user="u1")
    acc.measure_list(ItemList([2]), ItemList([2]), user="u2")

    df = acc.list_metrics()
    summary = acc.summary_metrics()

    assert "Mixed.mean" in summary
    assert summary["Mixed.mean"] == 1.5
    assert len(df) == 2


def test_none_extract_metric():
    class NoneExtractMetric(Metric):
        label = "none_extract"

        def measure_list(self, output, test):
            return 4

        def extract_list_metrics(self, data):
            return None

        def summarize(self, values):
            return {"mean": sum(values) / len(values)}

    acc = MetricAccumulator()
    acc.add_metric(NoneExtractMetric())
    acc.measure_list(ItemList([1]), ItemList([1]), user="u1")
    df = acc.list_metrics()
    assert pd.isna(df.loc["u1", "none_extract"])


# metricWrapper properties and summarization


def test_metricwrapper_is_decomposed_property():
    class DummyDecomposed(DecomposedMetric):
        label = "dummy_decomp"

        def compute_list_data(self, recs, test):
            return {"a": 1.0}

        def global_aggregate(self, values):
            return {"mean": 1.0}

        def measure_list(self, recs, test):
            return {"a": 1.0}

        def summarize(self, values):
            return {"mean": 1.0}

    wrapper = MetricWrapper(DummyDecomposed(), "decomp")
    assert wrapper.is_decomposed
    wrapper_non = MetricWrapper(ListLength(), "len")
    assert not wrapper_non.is_decomposed


def test_measure_metric_with_none_summarize():
    """Test metric that returns None from summarize."""

    class NoSummarizeMetric(Metric):
        label = "no_summarize"

        def measure_list(self, recs, test):
            return 7.0

        def summarize(self, values):
            return None

    acc = MetricAccumulator()
    acc.add_metric(NoSummarizeMetric())
    acc.measure_list(ItemList([1]), ItemList([1]), user="u1")

    summary = acc.summary_metrics()
    assert "no_summarize" not in summary


# test with movielens data


def test_full_workflow_integration_improved(ml_ds):
    split = split_temporal_fraction(ml_ds, 0.2, filter_test_users=True)
    scorer = PopScorer()
    scorer.train(split.train)
    acc = MetricAccumulator()
    acc.add_metric(Recall(10))
    acc.add_metric(NDCG(10))

    all_items = split.train.items.ids()
    test_users = list(split.test)[:5]

    for user, truth_il in test_users:
        scores = scorer(ItemList(all_items, user=[user.user_id] * len(all_items)))
        recs_il = ItemList(scores.top_n(10), user=[user.user_id] * 10, ordered=True)
        acc.measure_list(recs_il, truth_il, user=user.user_id)

    list_metrics = acc.list_metrics()
    summary = acc.summary_metrics()

    assert len(list_metrics) == len(test_users)
    assert set(list_metrics.columns) == {"Recall@10", "NDCG@10"}
    assert all(0 <= list_metrics[col].max() <= 1 for col in list_metrics.columns)

    assert len(summary) > 0
    for key, value in summary.items():
        if value is not None:
            if "std" in key.lower():
                assert value >= 0
            else:
                assert 0 <= value <= 1


# test that global metric raises errors for unsupported operations


def test_global_metric_unsupported():
    class AnotherGlobalMetric(GlobalMetric):
        label = "global"

        def measure_run(self, run, test):
            return 1.0

    metric = AnotherGlobalMetric()

    with raises(NotImplementedError, match="Global metrics don't support per-list measurement"):
        metric.measure_list(ItemList([1, 2]), ItemList([1]))

    with raises(NotImplementedError, match="Global metrics should implement measure_run instead"):
        metric.summarize([1, 2, 3])


# test edge cases in Metric.summarize


def test_list_metric_summarize_edge_cases():
    class TestListMetric(ListMetric):
        def measure_list(self, output, test):
            return len(output)

    metric = TestListMetric()

    # with empty list
    result = metric.summarize([])
    assert result["mean"] is None
    assert result["median"] is None
    assert result["std"] is None

    # with PyArrow array input
    arr = pa.array([1.0, 2.0, 3.0])
    result = metric.summarize(arr)
    assert result["mean"] == 2.0
    assert result["median"] == 2.0
    assert result["std"] == 1.0


def test_decomposed_metric_numeric_return():
    class TestDecomposedMetric(DecomposedMetric):
        def compute_list_data(self, output, test):
            return len(output)

        def global_aggregate(self, values):
            return 5.0

    metric = TestDecomposedMetric()
    result = metric.summarize([1, 2, 3])
    assert result == {"value": 5.0}


def test_empty_intermediate_values():
    class TestMetric(Metric):
        label = "test"

        def measure_list(self, recs, test):
            return None  # no intermediate data

        def summarize(self, values):
            return {"mean": 0.0}

    acc = MetricAccumulator()
    acc.add_metric(TestMetric())

    summary = acc.summary_metrics()
    assert not summary or "test" not in summary


def test_accumulator_duplicate_labels():
    acc = MetricAccumulator()
    acc.add_metric(ListLength(), label="dup")
    acc.add_metric(ListLength(), label="dup")

    with raises(RuntimeError, match="duplicate metric"):
        acc._validate_setup()


def test_accumulator_with_default_value_on_none_result():
    class NoneMetric(Metric):
        label = "none_metric"

        def measure_list(self, recs, test):
            return None

        def summarize(self, values):
            return {"mean": 123}

    acc = MetricAccumulator()
    acc.add_metric(NoneMetric(), default=99.0)
    acc.measure_list(ItemList([1]), ItemList([1]), user="u1")
    metrics = acc.list_metrics()
    # none should fall back to default value
    assert metrics.loc["u1", "none_metric"] == 99.0



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd
import torch

from pytest import approx, raises

from lenskit.basic import BiasModel, BiasScorer
from lenskit.data import Dataset, from_interactions_df
from lenskit.data.items import ItemList
from lenskit.operations import predict, recommend
from lenskit.pipeline import Pipeline, PipelineBuilder, topn_pipeline
from lenskit.testing import BasicComponentTests, ScorerTests

_log = logging.getLogger(__name__)

simple_df = pd.DataFrame(
    {"item": [1, 1, 2, 3], "user": [10, 12, 10, 13], "rating": [4.0, 3.0, 5.0, 2.0]}
)
simple_ds = from_interactions_df(simple_df)


class TestBias(BasicComponentTests, ScorerTests):
    component = BiasScorer
    configs = [{"damping": 10}, {"damping": {"user": 5, "item": 25}}]
    can_score = "all"


def test_bias_check_arguments():
    # negative damping is not allowed
    with raises(ValueError):
        BiasScorer(damping=-1)

    # negative user damping not allowed
    with raises(ValueError):
        BiasScorer(damping=(-1, 5))

    # negative user damping not allowed
    with raises(ValueError):
        BiasScorer(damping=(5, -1))


def test_bias_partial(ml_ratings, rng):
    ml_ratings = ml_ratings.copy()
    ml_ratings.loc[rng.choice(len(ml_ratings), 1000, replace=False), "rating"] = np.nan
    ds = from_interactions_df(ml_ratings)

    bias = BiasScorer(damping=5)
    bias.train(ds)
    assert np.isfinite(bias.model_.global_bias)
    assert np.all(np.isfinite(bias.model_.item_biases))


def test_bias_full():
    bias = BiasModel.learn(simple_ds)
    assert bias.global_bias == approx(3.5)

    assert bias.items is not None
    assert set(bias.items.ids()) == set([1, 2, 3])
    off = bias.item_biases
    assert off is not None
    exp = pd.Series([0, 1.5, -1.5], index=[1, 2, 3])
    exp = exp.reindex(bias.items.ids())
    df = pd.DataFrame({"computed": off, "expected": exp}).join(
        simple_df.groupby("item")["rating"].mean()
    )
    _log.debug("item stats:\n%s", df)
    assert off == approx(exp.values)

    assert bias.users is not None
    assert set(bias.users.ids()) == set([10, 12, 13])
    off = bias.user_biases
    assert off is not None
    exp = pd.Series([0.25, -0.5, 0], index=[10, 12, 13])
    exp = exp.reindex(bias.users.ids())
    _log.debug("computed user offsets:\n%s", off)
    _log.debug("expected user offsets:\n%s", exp)
    assert off == approx(exp.values)


def test_bias_clone():
    bias = BiasScorer()
    bias.train(simple_ds)

    params = bias.dump_config()
    assert sorted(params.keys()) == ["damping", "entities"]

    a2 = BiasScorer(BiasScorer.validate_config(params))
    assert a2 is not bias
    assert getattr(a2, "model_", None) is None


def test_bias_clone_damping():
    bias = BiasScorer(damping={"user": 10, "item": 5})
    bias.train(simple_ds)

    params = bias.dump_config()
    assert sorted(params.keys()) == ["damping", "entities"]

    a2 = BiasScorer(BiasScorer.validate_config(params))
    assert a2 is not bias
    assert isinstance(a2.config.damping, dict)
    assert a2.config.damping["user"] == 10
    assert a2.config.damping["item"] == 5
    assert getattr(a2, "model_", None) is None


def test_bias_global_only():
    bias = BiasModel.learn(simple_ds, entities=[])
    assert bias.global_bias == approx(3.5)
    assert bias.items is None
    assert bias.item_biases is None
    assert bias.users is None
    assert bias.user_biases is None


def test_bias_no_user():
    bias = BiasModel.learn(simple_ds, entities={"item"})
    assert bias.global_bias == approx(3.5)

    assert bias.item_biases is not None
    assert bias.item_biases == approx(np.array([0, 1.5, -1.5]))

    assert bias.user_biases is None


def test_bias_no_item():
    bias = BiasModel.learn(simple_ds, entities={"user"})
    assert bias.global_bias == approx(3.5)
    assert bias.item_biases is None

    assert bias.user_biases is not None
    assert bias.user_biases == approx(np.array([1.0, -0.5, -1.5]))


def test_bias_global_predict():
    bias = BiasScorer(entities=[])
    bias.train(simple_ds)

    p = bias(10, ItemList(item_ids=[1, 2, 3]))

    assert len(p) == 3
    assert p.scores() == approx(bias.model_.global_bias)


def test_bias_item_predict():
    bias = BiasScorer(entities={"item"})
    bias.train(simple_ds)
    assert bias.model_.item_biases is not None

    p = bias(10, ItemList(item_ids=[1, 2, 3]))

    assert len(p) == 3
    assert p.scores() == approx((bias.model_.item_biases + bias.model_.global_bias))


def test_bias_user_predict():
    bias = BiasScorer(entities={"user"})
    bias.train(simple_ds)
    bm = bias.model_
    p = bias(10, ItemList(item_ids=[1, 2, 3]))

    assert len(p) == 3
    assert p.scores() == approx(bm.global_bias + 1.0)

    p = bias(12, ItemList(item_ids=[1, 3]))

    assert len(p) == 2
    assert p.scores() == approx(bm.global_bias - 0.5)


def test_bias_new_user_predict():
    bias = BiasScorer()
    bias.train(simple_ds)
    bm = bias.model_
    assert bm.item_biases is not None

    items = ItemList(item_ids=[1, 2, 3], rating=[1.5, 2.5, 3.5])
    p = bias(items, ItemList(item_ids=[1, 3]))

    ratings = items.field("rating")
    assert ratings is not None
    offs = ratings - bm.global_bias - bm.item_biases
    umean = offs.mean()
    _log.info("user mean is %f", umean)

    assert len(p) == 2
    assert p.scores() == approx((bm.global_bias + bm.item_biases + umean)[[0, 2]])


def test_bias_predict_unknown_item():
    bias = BiasScorer()
    bias.train(simple_ds)
    bm = bias.model_

    assert bm.items is not None
    assert bm.item_biases is not None

    p = bias(10, ItemList(item_ids=[1, 3, 4]))

    assert len(p) == 3
    locs = bm.items.numbers([1, 3])
    intended = bm.item_biases[locs] + bm.global_bias + 0.25
    ps = p.scores("pandas", index="ids")
    assert ps is not None
    assert ps.loc[[1, 3]].values == approx(intended)
    assert ps.loc[4] == approx(bm.global_bias + 0.25)


def test_bias_predict_unknown_user():
    bias = BiasScorer()
    bias.train(simple_ds)
    bm = bias.model_
    assert bm.items is not None
    assert bm.item_biases is not None

    p = bias(15, ItemList(item_ids=[1, 3]))

    assert len(p) == 2
    locs = bm.items.numbers([1, 3])
    assert p.scores() == approx((bm.item_biases[locs] + bm.global_bias))


def test_bias_train_ml_ratings(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    bias = BiasScorer()
    bias.train(ml_ds)
    bm = bias.model_
    assert bm.items is not None
    assert bm.item_biases is not None

    assert bm.global_bias == approx(ml_ratings.rating.mean())
    imeans_data = ml_ds.item_stats()["mean_rating"]
    imeans_algo = bm.item_biases + bm.global_bias
    ares, data = pd.Series(imeans_algo, index=bm.items.ids()).align(
        imeans_data, fill_value=bm.global_bias
    )
    assert ares.values == approx(data.values)

    urates = ml_ratings.set_index("user_id").loc[2].set_index("item_id").rating
    umean = (urates - imeans_data[urates.index]).mean()
    p = bias(2, ItemList(item_ids=[10, 11, -1]))
    assert len(p) == 3
    ps = p.scores("pandas", index="ids")
    assert ps is not None
    assert ps.iloc[0] == approx(imeans_data.loc[10] + umean)
    assert ps.iloc[1] == approx(imeans_data.loc[11] + umean)
    assert ps.iloc[2] == approx(ml_ratings.rating.mean() + umean)


def test_bias_item_damp():
    bias = BiasModel.learn(simple_ds, entities={"item"}, damping=5)
    assert bias.global_bias == approx(3.5)

    assert bias.item_biases is not None
    assert bias.item_biases == approx(np.array([0, 0.25, -0.25]))

    assert bias.user_biases is None


def test_bias_user_damp():
    bias = BiasModel.learn(simple_ds, entities={"user"}, damping=5)
    assert bias.global_bias == approx(3.5)
    assert bias.item_biases is None

    assert bias.user_biases is not None
    assert bias.user_biases == approx(np.array([0.2857, -0.08333, -0.25]), abs=1.0e-4)


def test_bias_damped():
    bias = BiasModel.learn(simple_ds, damping=5)
    assert bias.global_bias == approx(3.5)

    assert bias.item_biases is not None
    assert bias.item_biases == approx(np.array([0, 0.25, -0.25]))

    assert bias.user_biases is not None
    assert bias.user_biases == approx(np.array([0.25, -00.08333, -0.20833]), abs=1.0e-4)


def test_bias_separate_damping():
    bias = BiasModel.learn(simple_ds, damping=(5, 10))
    assert bias.global_bias == approx(3.5)

    assert bias.item_biases is not None
    assert bias.item_biases == approx(np.array([0, 0.136364, -0.13636]), abs=1.0e-4)

    assert bias.user_biases is not None
    assert bias.user_biases == approx(np.array([0.266234, -0.08333, -0.22727]), abs=1.0e-4)


def test_bias_save():
    original = BiasScorer(damping=5)
    original.train(simple_ds)
    assert original.model_.global_bias == approx(3.5)

    _log.info("saving baseline model")
    data = pickle.dumps(original)
    _log.info("serialized to %d bytes", len(data))

    copy = pickle.loads(data)

    assert copy.model_.global_bias == original.model_.global_bias
    assert np.all(copy.model_.item_biases == original.model_.item_biases)
    assert np.all(copy.model_.user_biases == original.model_.user_biases)


def test_bias_pipeline(ml_ds: Dataset):
    pipe = PipelineBuilder()
    user = pipe.create_input("user", int)
    items = pipe.create_input("items")

    bias = BiasScorer()
    bias.train(ml_ds)
    out = pipe.add_component("bias", bias, query=user, items=items)

    pipe = pipe.build()
    res = pipe.run(out, user=2, items=ItemList(item_ids=[10, 11, -1]))

    assert len(res) == 3
    assert np.all(res.ids() == [10, 11, -1])

    scores = res.scores()
    assert scores is not None
    assert not np.any(np.isnan(scores[:2]))


def test_bias_topn(ml_ds: Dataset):
    pipe = topn_pipeline(BiasScorer(), predicts_ratings=True, n=10)
    print(pipe.config)
    pipe.train(ml_ds)

    res = predict(pipe, 2, ItemList(item_ids=[10, 11, -1]))
    assert isinstance(res, ItemList)
    assert len(res) == 3
    assert np.all(res.ids() == [10, 11, -1])

    recs = recommend(pipe, 2, n=10)
    assert isinstance(recs, ItemList)
    assert len(recs) == 10


def test_bias_topn_run_length(ml_ds: Dataset):
    pipe = topn_pipeline(BiasScorer(), predicts_ratings=True, n=100)
    print(pipe.config)
    pipe.train(ml_ds)

    res = predict(pipe, 2, items=ItemList(item_ids=[10, 11, -1]))
    assert isinstance(res, ItemList)
    assert len(res) == 3
    assert np.all(res.ids() == [10, 11, -1])

    res = recommend(pipe, 2, n=10)
    assert isinstance(res, ItemList)
    assert len(res) == 10
